#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""This file implements the Win32 keyboard interface using sendinput."""

# pylint: disable=E0401
# This file imports Win32-only symbols.

from locale import getpreferredencoding
from struct import unpack
import time

from six import binary_type, text_type

import win32api
import win32con
import win32gui
import win32process

from ._base import BaseKeyboard, BaseTypeable, BaseKeySymbols
from ..sendinput import KeyboardInput, make_input_array, send_input_array


class Win32KeySymbols(BaseKeySymbols):
    """ Key symbols for win32. """

    # Whitespace and editing keys
    RETURN = win32con.VK_RETURN
    TAB = win32con.VK_TAB
    SPACE = win32con.VK_SPACE
    BACK = win32con.VK_BACK
    DELETE = win32con.VK_DELETE

    # Main modifier keys
    SHIFT = win32con.VK_SHIFT
    CONTROL = win32con.VK_CONTROL
    ALT = win32con.VK_MENU

    # Right modifier keys
    RSHIFT = win32con.VK_RSHIFT
    RCONTROL = win32con.VK_RCONTROL
    RALT = win32con.VK_RMENU

    # Special keys
    ESCAPE = win32con.VK_ESCAPE
    INSERT = win32con.VK_INSERT
    PAUSE = win32con.VK_PAUSE
    LSUPER = win32con.VK_LWIN
    RSUPER = win32con.VK_RWIN
    APPS = win32con.VK_APPS
    SNAPSHOT = win32con.VK_SNAPSHOT

    # Lock keys
    SCROLL_LOCK = win32con.VK_SCROLL
    NUM_LOCK = win32con.VK_NUMLOCK
    CAPS_LOCK = win32con.VK_CAPITAL

    # Navigation keys
    UP = win32con.VK_UP
    DOWN = win32con.VK_DOWN
    LEFT = win32con.VK_LEFT
    RIGHT = win32con.VK_RIGHT
    PAGE_UP = win32con.VK_PRIOR
    PAGE_DOWN = win32con.VK_NEXT
    HOME = win32con.VK_HOME
    END = win32con.VK_END

    # Number pad keys
    MULTIPLY = win32con.VK_MULTIPLY
    ADD = win32con.VK_ADD
    SEPARATOR = win32con.VK_SEPARATOR
    SUBTRACT = win32con.VK_SUBTRACT
    DECIMAL = win32con.VK_DECIMAL
    DIVIDE = win32con.VK_DIVIDE
    NUMPAD0 = win32con.VK_NUMPAD0
    NUMPAD1 = win32con.VK_NUMPAD1
    NUMPAD2 = win32con.VK_NUMPAD2
    NUMPAD3 = win32con.VK_NUMPAD3
    NUMPAD4 = win32con.VK_NUMPAD4
    NUMPAD5 = win32con.VK_NUMPAD5
    NUMPAD6 = win32con.VK_NUMPAD6
    NUMPAD7 = win32con.VK_NUMPAD7
    NUMPAD8 = win32con.VK_NUMPAD8
    NUMPAD9 = win32con.VK_NUMPAD9

    # Function keys
    F1 = win32con.VK_F1
    F2 = win32con.VK_F2
    F3 = win32con.VK_F3
    F4 = win32con.VK_F4
    F5 = win32con.VK_F5
    F6 = win32con.VK_F6
    F7 = win32con.VK_F7
    F8 = win32con.VK_F8
    F9 = win32con.VK_F9
    F10 = win32con.VK_F10
    F11 = win32con.VK_F11
    F12 = win32con.VK_F12
    F13 = win32con.VK_F13
    F14 = win32con.VK_F14
    F15 = win32con.VK_F15
    F16 = win32con.VK_F16
    F17 = win32con.VK_F17
    F18 = win32con.VK_F18
    F19 = win32con.VK_F19
    F20 = win32con.VK_F20
    F21 = win32con.VK_F21
    F22 = win32con.VK_F22
    F23 = win32con.VK_F23
    F24 = win32con.VK_F24

    # Multimedia keys
    VOLUME_UP = win32con.VK_VOLUME_UP
    VOLUME_DOWN = win32con.VK_VOLUME_DOWN
    VOLUME_MUTE = win32con.VK_VOLUME_MUTE
    MEDIA_NEXT_TRACK = win32con.VK_MEDIA_NEXT_TRACK
    MEDIA_PREV_TRACK = win32con.VK_MEDIA_PREV_TRACK
    MEDIA_PLAY_PAUSE = win32con.VK_MEDIA_PLAY_PAUSE
    BROWSER_BACK = win32con.VK_BROWSER_BACK
    BROWSER_FORWARD = win32con.VK_BROWSER_FORWARD

    # Include virtual-key codes for digits and letters defined here:
    # https://docs.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
    DIGITS_MAP = {chr(x): x for x in range(0x30, 0x39)}
    LOWERCASE_ALPHABET_MAP = {chr(x): x - 0x20 for x in range(0x61, 0x7b)}
    UPPERCASE_ALPHABET_MAP = {chr(x): x | 0x100 for x in range(0x41, 0x5b)}
    CHAR_VK_MAP = DIGITS_MAP.copy()
    CHAR_VK_MAP.update(LOWERCASE_ALPHABET_MAP)
    CHAR_VK_MAP.update(UPPERCASE_ALPHABET_MAP)


class Win32Typeable(BaseTypeable):

    __slots__ = ("_code", "_modifiers", "_name", "_is_text", "_char")

    def __init__(self, code, modifiers=(), name=None, is_text=False,
                 char=None):
        BaseTypeable.__init__(self, code, modifiers, name, is_text)
        self._char = char

    def update(self, hardware_events_required):
        # Nothing to do for virtual keys.
        if self._char is None:
            return True

        # Get updated key code and modifiers for this Typeable, falling back
        #  on the key codes in CHAR_VK_MAP where necessary.
        try:
            self._is_text = False
            code, modifiers = Win32Keyboard.get_keycode_and_modifiers(
                self._char, char_vk_fallback=hardware_events_required)
        except ValueError:
            if hardware_events_required:
                # This Typeable cannot be typed in the current context.
                return False
            else:
                # Fallback on Unicode events.
                code, modifiers = self._char, ()
                self._is_text = True

        # Set key code and modifiers.
        self._code, self._modifiers = code, modifiers
        return True

    def _unicode_events(self, down, timeout):
        character = self._code
        if not isinstance(character, text_type):
            raise TypeError("cannot create Unicode events for character: %r"
                            % character)

        # Construct the UTF-16 (LE) events that sendinput expects.
        events = []
        byte_stream = character.encode("utf-16-le")
        format_string = "<" + str(len(byte_stream) // 2) + "H"
        for short in unpack(format_string, byte_stream):
            if down is True or down is False:
                events.extend([(short, down, timeout, True)])
            else:  # both for events()
                events.extend([(short, True, timeout, True),
                               (short, False, timeout, True)])
        return events

    def on_events(self, timeout=0):
        """Return events for pressing this key down."""
        if self._is_text:
            events = self._unicode_events(True, timeout)
        else:
            events = [(m, True, 0) for m in self._modifiers]
            events.append((self._code, True, timeout))
        return events

    def off_events(self, timeout=0):
        """Return events for releasing this key."""
        if self._is_text:
            events = self._unicode_events(False, timeout)
        else:
            events = [(m, False, 0) for m in self._modifiers]
            events.append((self._code, False, timeout))
            events.reverse()
        return events

    def events(self, timeout=0):
        """Return events for pressing and then releasing this key."""
        if self._is_text:
            events = self._unicode_events("both", timeout)
        else:
            events = [(self._code, True, 0), (self._code, False, timeout)]
            for m in self._modifiers[-1::-1]:
                events.insert(0, (m, True, 0))
                events.append((m, False, 0))
        return events


class Win32Keyboard(BaseKeyboard):
    """Static class wrapper around SendInput."""

    shift_code = win32con.VK_SHIFT
    ctrl_code = win32con.VK_CONTROL
    alt_code = win32con.VK_MENU

    @classmethod
    def get_current_layout(cls):
        # Get the current window's keyboard layout.
        thread_id = win32process.GetWindowThreadProcessId(
            win32gui.GetForegroundWindow()
        )[0]
        return win32api.GetKeyboardLayout(thread_id)

    @classmethod
    def send_keyboard_events(cls, events):
        """
        Send a sequence of keyboard events.

        Positional arguments:
        events -- a sequence of tuples of the form
            (keycode, down, timeout), where
                keycode (int): Win32 Virtual Key code.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.
            or
            (character, down, timeout, is_text), where
                character (str): Unicode character.
                down (boolean): True means the key will be pressed down,
                    False means the key will be released.
                timeout (int): number of seconds to sleep after
                    the keyboard event.
                is_text (boolean): True means that the keypress is targeted
                    at a window or control that accepts Unicode text.
        """
        layout = cls.get_current_layout()

        # Process and send keyboard events.
        items = []
        for event in events:
            if len(event) == 3:
                # Pass scancode=0 to press keys via virtual-key codes
                #  instead of scancodes.
                keycode, down, timeout = event
                input_structure = KeyboardInput(keycode, down,
                                                # scancode=0,
                                                layout=layout)
            elif len(event) == 4 and event[3]:
                character, down, timeout = event[:3]
                input_structure = KeyboardInput(0, down, scancode=character)
            else:
                raise ValueError("invalid keyboard event tuple: %r" % event)
            items.append(input_structure)
            if timeout:
                array = make_input_array(items)
                items = []
                send_input_array(array)
                time.sleep(timeout)
        if items:
            array = make_input_array(items)
            send_input_array(array)
            if timeout: time.sleep(timeout)

    @classmethod
    def _get_initial_keycode(cls, char, char_vk_fallback=True):
        # Get the code for this character.
        layout = cls.get_current_layout()
        try:
            code = win32api.VkKeyScanEx(char, layout)
        except TypeError:
            code = -1

        # Fallback on the keycode in CHAR_VK_MAP, if applicable.
        #  Note: char_vk_fallback=False is used when typing these characters
        #  with the Unicode keyboard.
        if (char_vk_fallback and code == -1 and
                char in Win32KeySymbols.CHAR_VK_MAP):
            code = Win32KeySymbols.CHAR_VK_MAP[char]

        if code == -1:
            raise ValueError("Unknown char: %r" % char)

        return code

    @classmethod
    def xget_virtual_keycode(cls, char, char_vk_fallback=True):
        code = cls._get_initial_keycode(char, char_vk_fallback)

        # Construct a list of the virtual key code and modifiers.
        codes = [code & 0x00ff]
        if code & 0x0100: codes.append(cls.shift_code)
        if code & 0x0200: codes.append(cls.ctrl_code)
        if code & 0x0400: codes.append(cls.alt_code)
        return codes

    @classmethod
    def get_keycode_and_modifiers(cls, char, char_vk_fallback=True):
        code = cls._get_initial_keycode(char, char_vk_fallback)

        # Construct a list of the virtual key code and modifiers.
        modifiers = []
        if code & 0x0100: modifiers.append(cls.shift_code)
        if code & 0x0200: modifiers.append(cls.ctrl_code)
        if code & 0x0400: modifiers.append(cls.alt_code)
        code &= 0x00ff
        return code, modifiers

    @classmethod
    def get_typeable(cls, char, is_text=False):
        if isinstance(char, binary_type):
            char = char.decode(getpreferredencoding())
        if is_text:
            return Win32Typeable(char, is_text=True, char=char)

        # Get the key code and modifiers for the Typeable.
        code, modifiers = cls.get_keycode_and_modifiers(char)
        return Win32Typeable(code, modifiers, name=char, char=char)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This module initializes the keyboard interface for the current platform.
"""

import os
import sys

# TODO Implement classes for Wayland (XDG_SESSION_TYPE == "wayland").

if sys.platform.startswith("win"):
    # Import Win32 classes.
    from ._win32 import (
        Win32Keyboard as Keyboard,
        Win32Typeable as Typeable,
        Win32KeySymbols as KeySymbols
    )

elif sys.platform == "darwin":
    from ._pynput import (
        PynputKeyboard as Keyboard,
        PynputTypeable as Typeable,
        DarwinKeySymbols as KeySymbols)

elif os.environ.get("XDG_SESSION_TYPE") == "x11":
    # Import classes for X11.  This is typically used on Unix-like systems.
    # The XDG_SESSION_TYPE environment variable may not be set in some
    #  circumstances, in which case it can be set manually in ~/.profile.
    from ._x11_base import (
        X11Typeable as Typeable,
        XdoKeySymbols as KeySymbols
    )

    # Import the keyboard for typing through xdotool.
    from ._x11_xdotool import XdotoolKeyboard as Keyboard

    # The libxdo implementation doesn't work with Python 3, so it is not
    #  used.
    # from ._x11_libxdo import LibxdoKeyboard as Keyboard

else:
    # No keyboard implementation is available. Dragonfly can function
    #  without a keyboard class, so don't raise an error or log any
    #  messages.  Error messages will occur later if and when keyboard
    #  events are sent.
    from ._base import (BaseKeyboard as Keyboard,
                        BaseTypeable as Typeable,
                        MockKeySymbols as KeySymbols)

# Initialize a Keyboard instance.
keyboard = Keyboard()



# coding=utf-8
#
# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>
#
# Modified from Aenea's server_osx.py file.

"""
Window class for macOS
============================================================================

"""

# pylint: disable=E0401
# This file imports MacOS-only symbols.

# pylint: disable=W0622
# Suppress warnings about redefining the built-in 'id' function.

import locale
import logging
import os

from six import binary_type
import applescript

from .base_window import BaseWindow
from .rectangle import Rectangle


class DarwinWindow(BaseWindow):
    """
        The Window class is an interface to the macOS window control and
        placement.

    """

    _log = logging.getLogger("window")

    #-----------------------------------------------------------------------
    # Class methods to create new Window objects.

    @classmethod
    def get_foreground(cls):
        script = u'''
        global theId
        tell application "System Events"
            set theId to id of first application process whose ¬
            frontmost is true
        end tell
        return theId
        '''
        window_id = applescript.AppleScript(script).run()
        if isinstance(window_id, binary_type):
            window_id = window_id.decode(locale.getpreferredencoding())

        return cls.get_window(id=int(window_id))

    @classmethod
    def get_matching_windows(cls, executable=None, title=None):
        # Convert absolute executable paths to basenames on macOS. This is
        # necessary because DarwinWindow executable names are not absolute
        # and will therefore never match.
        if executable is not None and os.path.isabs(executable):
            executable = os.path.basename(executable)

        return super(DarwinWindow, cls).get_matching_windows(
            executable, title
        )

    @classmethod
    def get_all_windows(cls):
        script = u'''
        global appIds
        tell application "System Events"
            set appIds to id of every application process whose ¬
            background only is false
        end tell
        return appIds
        '''
        return [cls.get_window(int(app_id)) for app_id in
                applescript.AppleScript(script).run()]

    #-----------------------------------------------------------------------
    # Methods for initialization and introspection.

    def __init__(self, id):
        BaseWindow.__init__(self, id)
        self._names.add(str(id))

    #-----------------------------------------------------------------------
    # Methods and properties for window attributes.

    def get_properties(self):
        """
        Method to get the properties of a macOS window.

        :rtype: dict
        :returns: window properties
        """
        script = '''
        tell application "System Events" to tell application process id "%s"
            try
                get properties of window 1
            on error errmess
                log errmess
            end try
        end tell
        ''' % self._id
        properties = applescript.AppleScript(script).run()
        if not properties:
            return {}

        result = {}
        encoding = locale.getpreferredencoding()
        for key, value in properties.items():
            key = key.code
            if isinstance(key, binary_type):
                key = key.decode(encoding)
            if isinstance(value, applescript.AEType):
                value = value.code
                if isinstance(value, binary_type):
                    value = value.decode(encoding)
            result[key] = value
        return result

    def get_attribute(self, attribute):
        """
        Method to get an attribute of a macOS window.

        **Note**: this method doesn't distinguish between multiple instances
        of the same application.

        :param attribute: attribute name
        :type attribute: string
        :returns: attribute value
        """
        script = '''
        tell application "%s"
            try
                get %s of window 1
            on error errmess
                log errmess
            end try
        end tell
        ''' % (self.executable, attribute)
        return applescript.AppleScript(script).run()

    def _get_window_text(self):
        return self.get_properties().get('pnam', '')

    def _get_class_name(self):
        return self.get_properties().get('pcls', '')

    def _get_window_module(self):
        script = '''
        global module
        set module to ""
        try
            tell application "System Events" to tell application process id %s
                set module to name
            end tell
        end try
        return module
        ''' % (self._id)
        return applescript.AppleScript(script).run()

    def _get_window_pid(self):
        script = '''
        global pid
        set pid to -1
        try
            tell application "System Events" to tell application process id %s
                set pid to unix id
            end tell
        end try
        return pid
        ''' % (self._id)
        return applescript.AppleScript(script).run()

    @property
    def is_minimized(self):
        return self.get_attribute('miniaturized')

    @property
    def is_maximized(self):
        return self.get_attribute('zoomed')

    @property
    def is_visible(self):
        return self.get_attribute('visible')

    #-----------------------------------------------------------------------
    # Methods related to window geometry.

    def get_position(self):
        props = self.get_properties()
        return Rectangle(props['posn'][0], props['posn'][1],
                         props['ptsz'][0], props['ptsz'][1])

    def set_position(self, rectangle):
        assert isinstance(rectangle, Rectangle)
        script = '''
        tell application "System Events"
            set firstWindow to first window of application process id "%s"
            set position of firstWindow to {%d, %d}
            set size of firstWindow to {%d, %d}
        end tell
        ''' % (self._id, rectangle.x, rectangle.y, rectangle.dx,
               rectangle.dy)
        applescript.AppleScript(script).run()

    #-----------------------------------------------------------------------
    # Methods for miscellaneous window control.

    def _press_window_button(self, button_subrole, action):
        # Note: The negation symbol ¬ is used to split long lines in
        # AppleScript.
        script = u'''
        tell application "System Events"
            perform action "%s" of (first button whose subrole is "%s") of ¬
            first window of application process id "%s"
        end tell
        ''' % (action, button_subrole, self._id)
        try:
            applescript.AppleScript(script).run()
            return True
        except applescript.ScriptError as err:
            self._log.error("Failed to perform window button action "
                            "%s -> %s: %s", button_subrole, action, err)
            return False

    def minimize(self):
        return self._press_window_button("AXMinimizeButton", "AXPress")

    def maximize(self):
        return self._press_window_button("AXFullScreenButton",
                                         "AXZoomWindow")

    def full_screen(self):
        """
        Enable full screen mode for this window.

        **Note**: this doesn't allow transitioning out of full screen mode.
        """
        return self._press_window_button("AXFullScreenButton", "AXPress")

    def restore(self):
        # Toggle maximized/minimized state if necessary.
        if self.is_maximized:
            return self.maximize()

        if self.is_minimized:
            return self.minimize()

        return True

    def close(self):
        return self._press_window_button("AXCloseButton", "AXPress")

    def set_foreground(self):
        script = '''
        tell application "System Events" to tell application process id "%s"
            set frontmost to true
        end tell
        ''' % self._id
        applescript.AppleScript(script).run()

    def set_focus(self):
        # Setting window focus without raising the window doesn't appear to
        # be possible in macOS, so fallback on set_foreground().
        self.set_foreground()



"""
Tests for the CMU Pocket Sphinx engine

Most engine functionality is tested here, although the tests are done
entirely via `mimic`, so there are some things which have to be tested
manually.

These tests assume the US English pronunciation dictionary, acoustic and
language models distributed with the `pocketsphinx` Python package are used.
"""

import unittest

import logging

from dragonfly.engines import (EngineBase, EngineError, MimicFailure,
                               get_engine)
from dragonfly.grammar.elements import Literal, Sequence, ListRef
from dragonfly.grammar.list import List
from dragonfly.grammar.grammar_base import Grammar
from dragonfly.grammar.rule_base import Rule
from dragonfly.grammar.recobs import RecognitionObserver
from dragonfly.grammar.rule_compound import CompoundRule
from dragonfly.test import (ElementTester, RuleTestGrammar)


class MockLoggingHandler(logging.Handler):
    """
    Mock logging handler to check for expected logs.
    Adapted this from a Stack Overflow answer: https://stackoverflow.com/a/1049375
    """

    def __init__(self, *args, **kwargs):
        self.messages = {
            'debug': [],
            'info': [],
            'warning': [],
            'error': [],
            'critical': [],
        }
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        self.messages[record.levelname.lower()].append(record.getMessage())


class RecognitionObserverTester(RecognitionObserver):
    """ RecognitionObserver class from the recobs doctests. """

    def __init__(self):
        RecognitionObserver.__init__(self)
        self.waiting = False
        self.words = None

    def on_begin(self):
        self.waiting = True
        self.words = None

    def on_recognition(self, words):
        self.waiting = False
        self.words = words

    def on_failure(self):
        self.waiting = False
        self.words = False


class SphinxEngineCase(unittest.TestCase):
    """
    Base TestCase class for Sphinx engine tests
    """

    log = logging.getLogger("engine")
    compile_log = logging.getLogger("engine.compiler")

    def setUp(self):
        self.engine = get_engine("sphinx")

        # Ensure the relevant configuration values are used.
        self.engine.config.TRAINING_DATA_DIR = ""
        self.engine.config.START_ASLEEP = False
        self.engine.config.WAKE_PHRASE = "wake up"
        self.engine.config.SLEEP_PHRASE = "go to sleep"
        self.engine.config.START_TRAINING_PHRASE = "start training session"
        self.engine.config.END_TRAINING_PHRASE = "end training session"
        self.engine.config.LANGUAGE = "en"

        # Map for test functions
        self.test_map = {}

        # Connect the engine.
        self.engine.connect()

        # Register a recognition observer.
        self.test_recobs = RecognitionObserverTester()
        self.test_recobs.register()

    def tearDown(self):
        self.test_map.clear()
        self.engine.resume_recognition()
        self.test_recobs.unregister()
        self.engine.disconnect()

    # ---------------------------------------------------------------------
    # Methods for control-flow assertion.

    def get_test_function(self):
        """
        Create and return a test function used for testing whether
        key phrases or rules are processed correctly, insofar as they reach
        the correct processing method/function.

        Note that returned test functions accept variable arguments.
        :return: callable
        """
        def func(*_):
            # Function was reached
            try:
                self.test_map[id(func)] += 1
            except KeyError:
                # Ignore any key errors; if this function's id is not in
                # test_map, then it's not relevant to the currently running
                # test method.
                pass

        self.test_map[id(func)] = 0
        return func

    def assert_test_function_called(self, func, n):
        """
        Assert that a test function was called n times.
        :type func: callable
        :param n: number of times the test function should have been called
        """
        x = self.test_map[id(func)]
        self.assertEqual(x, n, "wrapped test function was called %d time(s) "
                               "instead of %d time(s)" % (x, n))

    def reset_test_functions(self):
        """
        Reset the test_map values for all test functions.
        """
        for key in self.test_map.keys():
            self.test_map[key] = 0

    # ---------------------------------------------------------------------
    # Methods for asserting recognition success or failure.

    def assert_mimic_success(self, *phrases):
        """
        Assert that the engine can successfully mimic a number of speech
        strings.
        """
        try:
            self.engine.mimic_phrases(*phrases)
        except MimicFailure:
            self.fail("MimicFailure caught")

    def assert_mimic_failure(self, *phrases):
        self.assertRaises(MimicFailure, self.engine.mimic_phrases, *phrases)

    def assert_recobs_result(self, waiting, words):
        self.assertEqual(self.test_recobs.waiting, waiting)
        self.assertEqual(self.test_recobs.words, words)


class EngineTests(SphinxEngineCase):
    """
    Tests for most of the engine's functionality.
    """
    def test_get_engine_sphinx_is_usable(self):
        """
        Verify that the sphinx engine is usable by testing that a simple
        rule is loaded correctly and works correctly.
        """
        engine = get_engine()
        assert engine.name == "sphinx"
        assert isinstance(self.engine, EngineBase)
        engine.speak("testing sphinx")

        # Test that a basic rule can be loaded and recognized.
        seq = Sequence([Literal("hello"), Literal("world")])
        tester = ElementTester(seq, engine=engine)
        results = tester.recognize("hello world")
        self.assertEqual(results, [u"hello", u"world"])

    def test_engine_config(self):
        """ Verify that engine configuration is validated correctly. """
        # Use START_ASLEEP=True for this test.
        self.engine.config.START_ASLEEP = True
        options = [
            "DECODER_CONFIG",
            "LANGUAGE",

            "START_ASLEEP",
            "WAKE_PHRASE",
            "WAKE_PHRASE_THRESHOLD",
            "SLEEP_PHRASE",
            "SLEEP_PHRASE_THRESHOLD",

            "TRAINING_DATA_DIR",
            "TRANSCRIPT_NAME",
            "START_TRAINING_PHRASE",
            "START_TRAINING_PHRASE_THRESHOLD",
            "END_TRAINING_PHRASE",
            "END_TRAINING_PHRASE_THRESHOLD",

            "CHANNELS",
            "RATE",
            "SAMPLE_WIDTH",
            "FRAMES_PER_BUFFER",
        ]

        class TestConfig(object):
            pass

        # Set TestConfig values using the engine config.
        original_config = self.engine.config
        for name in options:
            setattr(TestConfig, name, getattr(original_config, name))

        def set_config(value):
            self.engine.config = value

        # Test that options are set to default values if deleted.
        # Don't compare decoder config objects with assertEqual because they
        # aren't comparable.
        for name in options:
            delattr(TestConfig, name)
            set_config(TestConfig)
            self.assertTrue(hasattr(TestConfig, name),
                            "%s was not reset" % name)
            if name != "DECODER_CONFIG":
                self.assertEqual(getattr(TestConfig, name),
                                 getattr(original_config, name),
                                 "%s did not match" % name)

    def test_pause_resume_recognition(self):
        """ Verify that pause/resume recognition functionality works. """

        grammar = RuleTestGrammar("test1")
        grammar.add_rule(CompoundRule(name="r1", spec="hello world"))

        def assert_recognize_succeeds():
            results = grammar.recognize_node("hello world").words()
            assert results == ["hello", "world"]

        # Enter sleep mode.
        self.engine.pause_recognition()
        self.assertTrue(self.engine.recognition_paused)

        # Mimicking hello world should succeed when recognition is
        # paused, but it will not succeed when actually speaking.
        assert_recognize_succeeds()

        # Check that recognition still succeeds when recognition is
        # resumed again.
        self.engine.resume_recognition()
        self.assertFalse(self.engine.recognition_paused)
        assert_recognize_succeeds()

        # Test that mimicking wake and sleep phrases also works
        # correctly.
        self.assert_mimic_success("go to sleep")
        self.assertTrue(self.engine.recognition_paused)
        self.assert_mimic_success("wake up")
        self.assertFalse(self.engine.recognition_paused)

    def test_start_asleep(self):
        """ Verify that the START_ASLEEP config option works. """
        # config.START_ASLEEP is False for the tests by default, so test
        # that first.
        self.assertFalse(self.engine.recognition_paused)
        self.assert_mimic_success("go to sleep")

        # Now set it to True, restart the engine and test again.
        self.engine.config.START_ASLEEP = True
        self.engine.disconnect()
        self.engine.connect()
        self.assertTrue(self.engine.recognition_paused)
        self.assert_mimic_success("wake up")

    def test_keyphrases_and_recobs(self):
        """ Verify that observers are notified of keyphrase events. """
        test1 = self.get_test_function()
        test2 = self.get_test_function()

        # Register two key phrases
        self.engine.set_keyphrase("hello world", 1e-20, test1)
        self.engine.set_keyphrase("testing testing", 1e-30, test2)

        def assert_success():
            self.assert_mimic_success("hello world")
            self.assert_recobs_result(False, (u"hello", u"world"))
            self.assert_test_function_called(test1, 1)
            self.assert_mimic_success("testing testing")
            self.assert_recobs_result(False, (u"testing", u"testing"))
            self.assert_test_function_called(test2, 1)
            self.reset_test_functions()

        # Test that both key phrases can be mimicked and that the observer
        # was notified.
        assert_success()

        # Test that both key phrases can be mimicked in sleep mode.
        self.engine.pause_recognition()
        assert_success()
        self.engine.resume_recognition()

        # Test that removed key phrases no longer match
        self.engine.unset_keyphrase("hello world")
        self.assert_mimic_failure("hello world")
        self.assert_recobs_result(False, False)
        self.assert_test_function_called(test1, 0)
        self.engine.unset_keyphrase("testing testing")
        self.assert_mimic_failure("testing testing")
        self.assert_recobs_result(False, False)
        self.assert_test_function_called(test1, 0)

    def test_set_keyphrase_unknown_words(self):
        """ Verify that setting a keyphrase with an unknown word raises an error.
        """
        from dragonfly.engines.backend_sphinx.engine import UnknownWordError
        self.assertRaises(UnknownWordError, self.engine.set_keyphrase,
                          "notaword", 1e-20, lambda: None)

    def test_built_in_keyphrase_unknown_words(self):
        """ Verify that errors are logged for built-in keyphrases with unknown words.
        """
        # Test invalid built-in keyphrases.
        self.engine.config.WAKE_PHRASE = "wake up unknownword"
        self.engine.config.SLEEP_PHRASE = "aninvalid sleepphrase"
        self.engine.config.START_TRAINING_PHRASE = "another invalidphrase"
        self.engine.config.END_TRAINING_PHRASE = "end trainingsession"

        # Restart the engine manually to verify that errors are logged for
        # the keyphrases on connect().
        handler = MockLoggingHandler()
        self.log.addHandler(handler)
        try:
            self.engine.disconnect()
            self.engine.connect()
        finally:
            self.log.removeHandler(handler)

        # Check the logged messages. Each of the unknown words in all four
        # built-in keyphrases should be listed in separate messages.
        errors = handler.messages["error"]
        self.assertEqual(len(errors), 4)
        self.assertIn("unknownword", errors[0])
        self.assertIn("aninvalid", errors[1])
        self.assertIn("sleepphrase", errors[1])
        self.assertIn("invalidphrase", errors[2])
        self.assertIn("trainingsession", errors[3])

    def test_unknown_grammar_words(self):
        """ Verify that warnings are logged for grammars with unknown words.
        """
        grammar = Grammar("test")
        grammar.add_rule(CompoundRule(name="r1", spec="testing unknownword"))
        grammar.add_rule(CompoundRule(name="r2", spec="wordz|natlink"))

        # Test with a list too.
        lst = List("lst", ["anotherunknownword", "testing multiplewords"])
        grammar.add_rule(CompoundRule(name="r3", spec="<lst>",
                                      extras=[ListRef("lst", lst)]))

        # Catch log messages and make some assertions.
        handler = MockLoggingHandler()
        self.compile_log.addHandler(handler)
        try:
            grammar.load()
            self.assertTrue(grammar.loaded)

            # Check the logged messages.
            warnings = handler.messages["warning"]
            self.assertEqual(len(warnings), 1)
            self.assertNotIn("testing", warnings[0])
            unknown_words = ["natlink", "unknownword", "anotherunknownword",
                             "wordz", "multiplewords"]
            for word in unknown_words:
                self.assertIn(word, warnings[0])

            # Test that list updates also produce warnings.
            lst.extend(("hello", "onemoreunknownword"))
            self.assertEqual(len(warnings), 2)
            self.assertNotIn("hello", warnings[1])
            self.assertIn("onemoreunknownword", warnings[1])
        finally:
            self.compile_log.removeHandler(handler)
            grammar.unload()

    def test_reference_names_with_spaces(self):
        """ Verify that reference names with spaces are accepted. """
        lst = List("my list", ["test list"])
        grammar = Grammar("My dragonfly grammar")
        grammar.add_rule(CompoundRule(name="my rule", spec="test rule"))
        grammar.add_rule(Rule(element=ListRef("my list", lst),
                              exported=True))
        try:
            grammar.load()
            self.assert_mimic_success("test rule")
            self.assert_mimic_success("test list")
        finally:
            grammar.unload()

    def test_training_session(self):
        """ Verify that no recognition processing occurs when training. """
        # Set up a rule to "train".
        test = self.get_test_function()

        class TestRule(CompoundRule):
            spec = "test training"
            _process_recognition = test

        # Create and load a grammar with the rule.
        grammar = Grammar("test")
        grammar.add_rule(TestRule())
        grammar.load()
        try:
            # Start a training session.
            self.engine.start_training_session()

            # Test that mimic succeeds, no processing occurs, and the
            # observer is still notified of events.
            self.assert_mimic_success("test training")
            self.assert_test_function_called(test, 0)
            self.assert_recobs_result(False, (u"test", u"training"))

            # End the session and test again.
            self.engine.end_training_session()
            self.assert_mimic_success("test training")
            self.assert_test_function_called(test, 1)
            self.assert_recobs_result(False, (u"test", u"training"))
        finally:
            grammar.unload()


# ---------------------------------------------------------------------


if __name__ == '__main__':
    unittest.main()



# This file is part of Aenea
#
# Aenea is free software: you can redistribute it and/or modify it under
# the terms of version 3 of the GNU Lesser General Public License as
# published by the Free Software Foundation.
#
# Aenea is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Aenea.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (2014) Alex Roper
# Alex Roper <alex@aroper.net>

"""
Mock module to allow dragonfly to be imported on linux locally.
Heavily modified to allow more dragonfly functionality to work
regardless of operating system.
"""

from .actions import ActionBase


class MockBase(object):
    def __init__(self, *args, **kwargs):
        pass


class MockAction(ActionBase):
    """ Mock class for dragonfly actions. """
    def __init__(self, *args, **kwargs):
        ActionBase.__init__(self)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Mimic action
============================================================================


The :class:`Mimic` action mimics a single recognition.  This is useful for
repeating a single prerecorded or predefined voice-command.

This class could for example be used to open a new Windows Explorer window::

    action = Mimic("open", "windows", "explorer")
    action.execute()


A more in-depth example is given below in the class reference.


Mimic quirks
----------------------------------------------------------------------------

Some SR engine back-ends have confusing :meth:`engine.mimic` method
behavior.  See the engine-specific mimic method documentation in sections
under :ref:`RefEngines` for more information.


Class reference
----------------------------------------------------------------------------

"""
from six               import string_types
from .action_base      import ActionBase, ActionError
from ..engines         import get_engine


# ---------------------------------------------------------------------------

class Mimic(ActionBase):
    """
        Mimic recognition action.

        The constructor arguments are the words which will be mimicked.
        These should be passed as a variable argument list.  For example::

            action = Mimic("hello", "world", r"!\\exclamation-mark")
            action.execute()

        If an error occurs during mimicking the given recognition, then an
        *ActionError* is raised.  A common error is that the engine does
        not know the given words and can therefore not recognize them.
        For example, the following attempts to mimic recognition of *one
        single word* including a space and an exclamation-mark; this will
        almost certainly fail::

            Mimic("hello world!").execute()   # Will raise ActionError.

        The constructor accepts the optional *extra* keyword argument, and
        uses this to retrieve dynamic data from the extras associated with
        the recognition.  For example, this can be used as follows to
        implement dynamic mimicking::

            class ExampleRule(MappingRule):
                mapping  = {
                            "mimic recognition <text> [<n> times]":
                                Mimic(extra="text") * Repeat(extra="n"),
                           }
                extras   = [
                            IntegerRef("n", 1, 10),
                            Dictation("text"),
                           ]
                defaults = {
                            "n": 1,
                           }

        The example above will allow the user to speak **"mimic
        recognition hello world! 3 times"**, which would result in the
        exact same output as if the user had spoken **"hello world!"**
        three times in a row.

    """

    def __init__(self, *words, **kwargs):
        ActionBase.__init__(self)
        self._words = tuple(words)
        if "extra" in kwargs:
            self._extra = kwargs.pop("extra")
        else:
            self._extra = None

        # Set pretty printing string used by __str__ and __unicode__.
        self._str = u", ".join(repr(w) for w in self._words)

        # Make sure that all keyword arguments have been consumed.
        if kwargs:
            raise ActionError("Invalid arguments: %r"
                              % ", ".join(list(kwargs.keys())))

    def _execute(self, data=None):
        engine = get_engine()
        words = self._words

        # If an extra was given, retrieve the associated value from
        #  the *data* dict and append it to the static words.
        if self._extra:
            try:
                extra = data[self._extra]
            except KeyError:
                raise ActionError("No extra data available for extra %r"
                                  % self._extra)

            # Append the extra data to the static words depending on
            #  the type of the extra data object.
            if isinstance(extra, engine.DictationContainer):
                words += tuple(extra.words)
            elif isinstance(extra, (tuple, list)):
                words += tuple(extra)
            elif isinstance(extra, string_types):
                words += (extra,)
            else:
                raise ActionError("Invalid extra data type: %r" % extra)

        # Mimic the series of words.
        self._log.debug("Mimicking recognition: %r", words)
        try:
            engine.disable_recognition_observers()
            engine.mimic(words)
            engine.enable_recognition_observers()
        except Exception as e:
            raise ActionError("Mimicking failed: %s" % e)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import os.path
import re
from setuptools import setup, find_packages, Command


#---------------------------------------------------------------------------
# Gather version from distribution file.

directory = os.path.abspath(os.path.dirname(__file__))
path = os.path.join(directory, "version.txt")
version_string = open(path).readline()
match = re.match(r"\s*(?P<rel>(?P<ver>\d+\.\d+)(?:\.\S+)*)\s*", version_string)
version = match.group("ver")
release = match.group("rel")

#---------------------------------------------------------------------------
# Override the 'test' command to use pytest instead.
# Test requirements are located in the 'test_requirements.txt' file.

class test(Command):
    description = 'run unit tests and doctests after in-place build'
    user_options = [
        # (long option, short option, description)
        # '=' means an argument should be supplied.
        ('test-suite=', None, 'Dragonfly engine to test (default: "text")'),
        ('pytest-options=', 'o',
            'pytest options (ex: "-s" to expose stdout/stdin)'),
    ]

    def initialize_options(self):
        self.test_suite = 'text'
        self.pytest_options = ''

    def finalize_options(self):
        # Check that 'test_suite' is an engine name.
        from dragonfly.test.suites import engine_tests_dict
        suite = self.test_suite
        assert suite in engine_tests_dict.keys(), \
            "the test suite value must be an engine name, not '%s'" % suite

        # Split pytest options into a list.
        self.pytest_options = self.pytest_options.split()

    def run(self):
        from dragonfly.test.suites import run_pytest_suite
        print("Test suite running for engine '%s'" % self.test_suite)
        result = run_pytest_suite(self.test_suite, self.pytest_options)

        # Exit using pytest's return code.
        exit(int(result))


#---------------------------------------------------------------------------
# Set up package.

def read(*names):
    return open(os.path.join(os.path.dirname(__file__), *names)).read()


setup(
      name             = "dragonfly2",
      version          = release,
      description      = "Speech recognition extension library",
      author           = "Christo Butcher",
      author_email     = "dist.dragonfly@twizzy.biz",
      maintainer       = "Dane Finlay",
      maintainer_email = "Danesprite@posteo.net",
      license          = "LICENSE.txt",
      url              = "https://github.com/dictation-toolbox/dragonfly",
      zip_safe         = False,  # To unzip documentation files.
      long_description = read("README.rst"),
      include_package_data=True,
      install_requires=[
                        "setuptools >= 40.0.0",
                        "packaging >= 19.0",
                        "six",
                        "pyperclip >= 1.7.0",
                        "enum34;python_version<'3.4'",
                        "regex",
                        "decorator",
                        "lark-parser == 0.8.*",

                        # Windows-only dependencies.
                        "comtypes;platform_system=='Windows'",
                        "pywin32;platform_system=='Windows'",

                        # Linux dependencies.
                        # "python-libxdo;platform_system=='Linux'",
                        # "Xlib;platform_system=='Linux'",
                        "psutil >= 5.5.1;platform_system=='Linux'",
                        "pynput >= 1.4.2;platform_system=='Linux'",

                        # Mac OS dependencies.
                        "pynput >= 1.4.2;platform_system=='Darwin'",
                        "pyobjc >= 5.2;platform_system=='Darwin'",
                        "py-applescript == 1.0.0;platform_system=='Darwin'",

                        # RPC requirements
                        "json-rpc",
                        "Werkzeug",
                        "requests",
                       ],

      extras_require={
          "sphinx": [
                     "sphinxwrapper >= 1.2.0",
                     "pyjsgf >= 1.7.0",
                     "pyaudio"
                    ],
          "kaldi": [
                    "kaldi-active-grammar ~= 2.1.0",
                    "sounddevice == 0.3.*",
                    "webrtcvad-wheels == 2.0.*",
                   ],
      },

      cmdclass={
          "test": test,
      },

      classifiers=[
                   "Development Status :: 4 - Beta",
                   "Environment :: Win32 (MS Windows)",
                   "Environment :: X11 Applications",
                   "Environment :: MacOS X",
                   "License :: OSI Approved :: "
                   "GNU Library or Lesser General Public License (LGPL)",
                   "Intended Audience :: Developers",
                   "Operating System :: Microsoft :: Windows",
                   "Operating System :: POSIX",
                   "Operating System :: MacOS",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3.4",
                   "Programming Language :: Python :: 3.5",
                   "Programming Language :: Python :: 3.6",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: Implementation :: CPython",
                   "Topic :: Multimedia :: Sound/Audio :: Speech",
                   "Topic :: Software Development :: Libraries :: Python Modules",
                  ],

      packages=find_packages(),

      test_suite="dragonfly.test.suites.text_suite",
     )



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
    This module demonstrates the use of Dragonfly's CompoundRule class.

    It shows how to use Dragonfly's Grammar, AppContext, and CompoundRule
    classes.  This module can be activated in the same way as other
    Natlink macros by placing it in the "NatLink/MacroSystem" folder.

"""

from dragonfly import (Grammar, AppContext, CompoundRule, Choice, Dictation)


#---------------------------------------------------------------------------
# Create this module's grammar and the context under which it'll be active.

grammar_context = AppContext(executable="notepad")
grammar = Grammar("notepad_example_2", context=grammar_context)


#---------------------------------------------------------------------------
# Create a compound rule which demonstrates CompoundRule and Choice types.

class FoodGroupRule(CompoundRule):

    spec   = "(I ate <food> <time> | <time> I ate <food>) [and thought it was <opinion>]"
    time   = {
              "(two days ago | day before yesterday)":  2,
              "yesterday":                              1,
              "today":                                  0,
             }
    food   = {
              # Note: quoted words are sometimes required for Dragon to
              # recognize commands.

              # Quoted food words.
              # Uncomment these mappings if Dragon isn't recognizing the
              # 'food' extra.
              # "an orange":                      "fruit",
              # '("a Granny Smith" | an) apple':  "fruit",
              # '"a hamburger"':                  "meat",
              # '"a steak"':                      "meat",
              # '"a juicy steak"':                "meat",

              # Unquoted food words.
              # Comment these mappings if using the quoted words.
              "an orange":                      "fruit",
              "(a Granny Smith | an) apple":    "fruit",
              "an orange":                      "fruit",
              "a hamburger":                    "meat",
              "a [juicy] steak":                "meat",
             }
    extras = [
              Choice("time", time),
              Choice("food", food),
              Dictation("opinion"),
             ]

    def _process_recognition(self, node, extras):
        days_ago  = extras["time"]
        foodgroup = extras["food"]
        print("You ate %s %d days ago." % (foodgroup, days_ago))
        if "opinion" in extras:
            print("You thought it was %s." % (extras["opinion"]))

grammar.add_rule(FoodGroupRule())


#---------------------------------------------------------------------------
# Load the grammar instance and define how to unload it.

grammar.load()

# Unload function which will be called by natlink at unload time.
def unload():
    global grammar
    if grammar: grammar.unload()
    grammar = None



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
"""
from dragonfly import Context

# ==========================================================================


class TestError(Exception):
    pass


# ==========================================================================

class Unique(object):
    """ Token class representing a unique identity. """

    def __init__(self, name):
        self._name = name

    def __repr__(self):
        return self._name


RecognitionFailure = Unique("RecognitionFailure")


# ==========================================================================


class TestContext(Context):
    """ Simple context class used to test rule and grammar contexts. """
    def __init__(self, active):
        super(TestContext, self).__init__()
        self.active = active

    def matches(self, executable, title, handle):
        # Ignore the parameters and return self.active.
        return self.active



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import logging
import os
import sys

import pytest

from dragonfly.log import setup_log
from dragonfly import get_engine

# Setup logging.
_log = logging.getLogger("dfly.test")
setup_log()


# ==========================================================================

common_names = [
    "test_accessibility",
    "test_actions",
    "test_contexts",
    "test_basic_rule",
    "test_engine_nonexistent",
    "test_log",
    "test_parser",
    "test_lark_parser",
    "test_rpc",
    "test_timer",
    "test_window",
    "documentation/test_action_base_doctest.txt",
    "documentation/test_grammar_elements_basic_doctest.txt",
    "documentation/test_grammar_elements_compound_doctest.txt",
    "documentation/test_grammar_list_doctest.txt",
    "documentation/test_recobs_doctest.txt",
]

# Run clipboard tests if on a desktop system.
desktop = (
    # Windows
    os.name == "nt" or
    # X11
    os.environ.get("XDG_SESSION_TYPE") == "x11" or
    # macOS
    sys.platform == "darwin"
)
if desktop:
    common_names.insert(2, "test_clipboard")

# Define spoken language test files. All of them work with the natlink and
# text engines. The English tests should work with sapi5 and sphinx by
# default.
language_names = [
    "test_language_de_number",
    "test_language_en_number",
    "test_language_nl_number",
]

# Use different test names depending on the DNS version. Fallback on v11 if
# the version is unknown.
try:
    import natlinkstatus
    dns_version = int(natlinkstatus.NatlinkStatus().getDNSVersion())
except:
    # Couldn't get the DNS version for whatever reason.
    dns_version = None

natlink_names = common_names + language_names + [
    "test_compiler_natlink",
    "test_dictation",
    "test_engine_natlink",
]

if dns_version and dns_version <= 10:
    natlink_names += ["documentation/test_word_formatting_v10_doctest.txt"]
else:
    natlink_names += ["documentation/test_word_formatting_v11_doctest.txt"]


# Define doctests for each engine.
engine_tests_dict = {
    "sapi5": [
        "test_engine_sapi5",
        "test_language_en_number",
    ] + common_names,

    "sphinx": [
        "test_engine_sphinx",
        "test_language_en_number",
        "test_dictation",
    ] + common_names,

    "kaldi": [
        "test_engine_kaldi",
        # "test_language_en_number",
    ] + common_names,

    "text": [
        "test_engine_text",
        "test_dictation",
    ] + common_names + language_names,

    "natlink": natlink_names,
}


# Exclude doctests that don't work with specific engines.
recobs_doctest = "documentation/test_recobs_doctest.txt"
lists_doctest = "documentation/test_grammar_list_doctest.txt"
engine_tests_dict['sapi5'].remove(recobs_doctest)
# engine_tests_dict['kaldi'].remove(recobs_doctest)
# engine_tests_dict['kaldi'].remove(lists_doctest)


# Add aliases of 'sapi5'.
engine_tests_dict['sapi5inproc'] = engine_tests_dict['sapi5']
engine_tests_dict['sapi5shared'] = engine_tests_dict['sapi5']


engine_params_dict = {
    "kaldi": dict(audio_input_device=False),
}


# ==========================================================================


def run_pytest_suite(engine_name, pytest_options):
    # Get test file paths.
    paths = []
    for name in engine_tests_dict[engine_name]:
        if name.startswith("test_"):
            # Use full module paths so pytest can import the files
            # correctly.
            name = "dragonfly/test/" + name + ".py"

        paths.append(name)

    # Initialize the engine and call connect().
    engine_params = engine_params_dict.get(engine_name, {})
    engine = get_engine(engine_name, **engine_params)
    engine.connect()

    # Prevent the engine from running timers on its own. This lets us
    # avoid race conditions.
    engine._timer_manager.disable()
    try:
        # Run doctests through pytest.main() now that the engine is set up.
        # Pass any specified pytest options, followed by doctest options for
        # compatibility with both Python 2.7 and 3.
        args = pytest_options + [
            '-o',
            'doctest_optionflags=ALLOW_UNICODE IGNORE_EXCEPTION_DETAIL'
        ] + paths
        return pytest.main(args)
    finally:
        # Disconnect after the tests.
        engine.disconnect()

        # Check that the dragonfly engine was not changed during the tests.
        assert engine is get_engine(), \
            "The registered engine changed during the test suite."



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Grammar class
============================================================================

"""

import logging
from six import string_types

from ..engines         import get_engine
from .rule_base        import Rule
from .list             import ListBase
from .context          import Context
from ..error           import GrammarError


# --------------------------------------------------------------------------

class Grammar(object):
    """
        Grammar class for managing a set of rules.

        This base grammar class takes care of the communication
        between Dragonfly's object model and the backend speech
        recognition engine.  This includes compiling rules and
        elements, loading them, activating and deactivating
        them, and unloading them.  It may, depending on the
        engine, also include receiving recognition results and
        dispatching them to the appropriate rule.

         - *name* -- name of this grammar
         - *description* (str, default: None) --
           description for this grammar
         - *context* (Context, default: None) --
           context within which to be active.  If *None*, the grammar will
           always be active.

    """

    # pylint: disable=too-many-instance-attributes
    _log_load     = logging.getLogger("grammar.load")
    _log_begin    = logging.getLogger("grammar.begin")
    _log_results  = logging.getLogger("grammar.results")
    _log          = logging.getLogger("grammar")

    # ----------------------------------------------------------------------
    # Methods for initialization and cleanup.

    def __init__(self, name, description=None, context=None, engine=None):
        self._name = name
        self._description = description
        if not (isinstance(context, Context) or context is None):
            raise TypeError("context must be either a Context object or "
                            "None")
        self._context = context

        if engine:
            self._engine = engine
        else:
            self._engine = get_engine()

        self._rules = []
        self._lists = []
        self._rule_names = None
        self._loaded = False
        self._enabled = True
        self._in_context = False

    def __del__(self):
        try:
            if self._loaded:
                return
            self.unload()
        except Exception as e:
            try:
                self._log.exception("Exception during grammar unloading:"
                                    " %s", e)
            except Exception as e:
                pass

    # ----------------------------------------------------------------------
    # Methods for runtime introspection.

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._name)

    name = property(lambda self: self._name,
                    doc="A grammar's name.")

    rules = property(lambda self: tuple(self._rules),
                     doc="List of a grammar's rules.")

    active_rules = property(
        lambda self: [r for r in self.rules if r.active],
        doc="List of a grammar's active rules."
    )

    lists = property(lambda self: tuple(self._lists),
                     doc="List of a grammar's lists.")

    loaded = property(lambda self: self._loaded,
                      doc="Whether a grammar is loaded into"
                          " its SR engine or not.")

    @property
    def rule_names(self):
        """
        List of grammar's rule names.
        """
        result = []
        for rule in self._rules:
            result.append(rule.name)
        return result

    def enable(self):
        """
            Enable this grammar so that it is active to receive
            recognitions.

        """
        self._enabled = True

    def disable(self):
        """
            Disable this grammar so that it is not active to
            receive recognitions.

        """
        self._enabled = False

    enabled = property(lambda self: self._enabled,
                       doc="Whether a grammar is active to receive "
                           "recognitions or not.")

    def set_exclusiveness(self, exclusive):
        """ Set the exclusiveness of this grammar. """
        self._engine.set_exclusiveness(self, exclusive)

    def set_exclusive(self, exclusive):
        """ Alias of :meth:`set_exclusiveness`. """
        self.set_exclusiveness(exclusive)

    def _set_engine(self, engine):
        if self._loaded:
            raise GrammarError(" Grammar %s: Cannot set engine while "
                               "loaded." % self)
        self._engine = engine

    engine = property(lambda self: self._engine, _set_engine,
                      doc="A grammar's SR engine.")

    def set_context(self, context):
        """
            Set the context for this grammar, under which it and its rules
            will be active and receive recognitions if it is also enabled.

            Use of this method overwrites any previous context.

            Contexts can be modified at any time, but will only be checked
            when :meth:`process_begin` is called.

            :param context: context within which to be active.  If *None*,
                the grammar will always be active.
            :type context: Context|None
        """
        if not (isinstance(context, Context) or context is None):
            raise TypeError("context must be either a Context object or "
                            "None")
        self._context = context

    context = property(lambda self: self._context,
                       doc="A grammar's context, under which it and its "
                           "rules will be active and receive recognitions "
                           "if it is also enabled.")

    # ----------------------------------------------------------------------
    # Methods for populating a grammar object instance.

    def add_rule(self, rule):
        """
        Add a rule to this grammar.

        The following rules apply when adding rules into grammars:

        #. Rules **cannot** be added to grammars that are currently loaded.
        #. Two or more rules with the same name are **not** allowed.

        .. warning::

           Note that while adding the same ``Rule`` object to more than one
           grammar is allowed, it is **not** recommended! This is because
           the context and active/enabled states of these rules will not
           function correctly if used. It is better to use *separate*
           ``Rule`` instances for each grammar instead.

        :param rule: Dragonfly rule
        :type rule: Rule
        """
        self._log_load.debug("Grammar %s: adding rule %s.",
                             self._name, rule.name)

        # Check for correct type and duplicate rules or rule names.
        if self._loaded:
            raise GrammarError("Cannot add rule while loaded.")
        elif not isinstance(rule, Rule):
            raise GrammarError("Invalid rule object: %s" % rule)
        elif rule in self._rules:
            return
        elif rule.imported:
            return
        elif [True for r in self._rules if r.name == rule.name]:
            raise GrammarError("Two rules with the same name '%s' not"
                               " allowed." % rule.name)
        elif rule.grammar is not None and rule.exported:
            self._log_load.warning("Exported rule %s is already in grammar "
                                   "%s, adding it to grammar %s is not "
                                   "recommended.", rule.name,
                                   rule.grammar.name, self._name)

        # Append the rule to this grammar object's internal list.
        self._rules.append(rule)
        rule.grammar = self

    def remove_rule(self, rule):
        """
        Remove a rule from this grammar.

        Rules **cannot** be removed from grammars that are currently loaded.

        :param rule: Dragonfly rule
        :type rule: Rule
        """
        self._log_load.debug("Grammar %s: removing rule %s.",
                             self._name, rule.name)

        # Check for correct type.
        if self._loaded:
            raise GrammarError("Cannot remove rule while loaded.")
        elif not isinstance(rule, Rule):
            raise GrammarError("Invalid rule object: %s" % rule)
        elif rule not in self._rules:
            return

        # Remove the rule from this grammar object's internal list.
        self._rules.remove(rule)
        rule.grammar = None

    def add_list(self, lst):
        """
        Add a list to this grammar.

        Lists **cannot** be added to grammars that are currently loaded.

        :param lst: Dragonfly list
        :type lst: ListBase
        """
        self._log_load.debug("Grammar %s: adding list %s.",
                             self._name, lst.name)

        # Make sure that the list can be loaded and is not a duplicate.
        if self._loaded:
            raise GrammarError("Cannot add list while loaded.")
        elif not isinstance(lst, ListBase):
            raise GrammarError("Invalid list object: %s" % lst)

        for l in self._lists:
            if l.name == lst.name:
                if l is lst:
                    # This list was already added previously, so ignore.
                    return
                raise GrammarError("Two lists with the same name '%s' not"
                                   " allowed." % lst.name)

        # Append the list to this grammar object's internal list.
        self._lists.append(lst)
        lst.grammar = self

    def remove_list(self, lst):
        """
        Remove a list from this grammar.

        Lists **cannot** be removed from grammars that are currently loaded.

        :param lst: Dragonfly list
        :type lst: ListBase
        """
        self._log_load.debug("Grammar %s: removing list %s.",
                             self._name, lst.name)

        # Check for correct type.
        if self._loaded:
            raise GrammarError("Cannot remove list while loaded.")
        elif not isinstance(lst, ListBase):
            raise GrammarError("Invalid list object: %s" % lst)
        elif lst.name not in [l.name for l in self._lists]:
            return

        # Remove the list from this grammar object's internal list.
        self._lists.remove(lst)
        lst.grammar = None

    def add_dependency(self, dep):
        """
            Add a rule or list dependency to this grammar.

            **Internal:** this method is normally *not* called
            by the user, but instead automatically during
            grammar compilation.

        """
        if isinstance(dep, Rule):
            self.add_rule(dep)
        elif isinstance(dep, ListBase):
            self.add_list(dep)
        else:
            raise GrammarError("Unknown dependency type %s." % dep)

    def add_all_dependencies(self):
        """
            Iterate through the grammar's rules and add all the necessary dependencies.

            **Internal** This method is called when the grammar is loaded.
        """
        memo = set()
        for r in self._rules:
            for d in r.dependencies(memo):
                self.add_dependency(d)

    # ----------------------------------------------------------------------
    # Methods for runtime modification of a grammar's contents.

    def activate_rule(self, rule):
        """
            Activate a rule loaded in this grammar.

            **Internal:** this method is normally *not* called
            directly by the user, but instead automatically when
            the rule itself is activated by the user.

        """
        self._log_load.debug("Grammar %s: activating rule %s.",
                             self._name, rule.name)

        # Check for correct type and valid rule instance.
        assert self._loaded
        assert isinstance(rule, Rule), \
            "Dragonfly rule objects must be of the type dragonfly.rule.Rule"
        if rule not in self._rules:
            raise GrammarError("Rule '%s' not loaded in this grammar."
                               % rule.name)
        if not rule.exported:
            return

        # Activate the given rule.
        self._engine.activate_rule(rule, self)

    def deactivate_rule(self, rule):
        """
            Deactivate a rule loaded in this grammar.

            **Internal:** this method is normally *not* called
            directly by the user, but instead automatically when
            the rule itself is deactivated by the user.

        """
        self._log_load.debug("Grammar %s: deactivating rule %s.",
                             self._name, rule.name)

        # Check for correct type and valid rule instance.
        assert self._loaded
        assert isinstance(rule, Rule), \
            "Dragonfly rule objects must be of the type dragonfly.rule.Rule"
        if rule not in self._rules:
            raise GrammarError("Rule '%s' not loaded in this grammar."
                               % rule.name)
        if not rule.exported:
            return

        # Deactivate the given rule.
        self._engine.deactivate_rule(rule, self)

    def update_list(self, lst):
        """
            Update a list's content loaded in this grammar.

            **Internal:** this method is normally *not* called
            directly by the user, but instead automatically when
            the list itself is modified by the user.

        """
        self._log_load.debug("Grammar %s: updating list %s.",
                             self._name, lst.name)

        # Check for correct type and valid list instance.
        #        assert self._loaded
        if lst not in self._lists:
            raise GrammarError("List '%s' not loaded in this grammar."
                               % lst.name)
        elif [True for w in lst.get_list_items()
              if not isinstance(w, string_types)]:
            raise GrammarError("List '%s' contains objects other than"
                               "strings." % lst.name)

        self._engine.update_list(lst, self)

    # ----------------------------------------------------------------------
    # Methods for registering a grammar object instance in natlink.

    def load(self):
        """ Load this grammar into its SR engine. """

        self._log_load.debug("Grammar %s: loading into engine %s.",
                             self._name, self._engine)

        # Prevent loading the same grammar multiple times.
        if self._loaded:
            return

        self.add_all_dependencies()
        self._engine.load_grammar(self)
        self._loaded = True
        self._in_context = False

        # Update all rules loaded in this grammar.
        for rule in self._rules:
            # Explicitly compare to False so that uninitialized rules (which
            # have active set to None) are activated.
            if rule.active is not False:
                rule.activate(force=True)
        # Update all lists loaded in this grammar.
        for lst in self._lists:
            # pylint: disable=protected-access
            lst._update()

        #        self._log_load.warning(self.get_complexity_string())

    def unload(self):
        """ Unload this grammar from its SR engine. """

        # Prevent unloading the same grammar multiple times.
        if not self._loaded:
            return
        self._log_load.debug("Grammar %s: unloading.", self._name)

        self._engine.unload_grammar(self)
        self._loaded = False
        self._in_context = False

    def get_complexity_string(self):
        """
            Build and return a human-readable text giving insight into the
            complexity of this grammar.

        """
        rules_all = self.rules
        rules_top = [r for r in self.rules if r.exported]
        rules_imp = [r for r in self.rules if r.imported]
        elements = []
        for rule in rules_all:
            elements.extend(self._get_element_list(rule))
        text = ("Grammar: %3d (%3d, %3d) rules, %4d elements (%3d avg)     %s"
                % (
                    len(rules_all), len(rules_top), len(rules_imp),
                    len(elements), len(elements) / len(rules_all),
                    self,
                )
                )
        for rule in rules_all:
            elements = self._get_element_list(rule)
            text += "\n  Rule: %4d  %s" % (len(elements), rule)
        return text

    def _get_element_list(self, thing):
        if isinstance(thing, Rule):
            element = thing.element
        else:
            element = thing
        elements = [element]
        for child in element.children:
            elements.extend(self._get_element_list(child))
        return elements

    # ----------------------------------------------------------------------
    # Callback methods for handling utterances and recognitions.

    def process_begin(self, executable, title, handle):
        """
            Start of phrase callback.

            *Usually derived grammar classes override
            ``Grammar._process_begin`` instead of this method, because
            this method merely wraps that method adding context matching.*

            This method is called when the speech recognition
            engine detects that the user has begun to speak a
            phrase.

            Arguments:
             - *executable* -- the full path to the module whose
               window is currently in the foreground.
             - *title* -- window title of the foreground window.
             - *handle* -- window handle to the foreground window.

        """
        # pylint: disable=expression-not-assigned

        self._log_begin.debug("Grammar %s: detected beginning of "
                              "utterance.", self._name)
        self._log_begin.debug("Grammar %s: executable '%s', title '%s'.",
                              self._name, executable, title)

        if not self._enabled:
            # Grammar is disabled, so deactivate all active rules.
            [r.deactivate() for r in self._rules if r.active]

        elif not self._context \
                or self._context.matches(executable, title, handle):
            # Grammar is within context.
            if not self._in_context:
                self._in_context = True
                self.enter_context()
            self._process_begin(executable, title, handle)
            for r in self._rules:
                if r.exported and hasattr(r, "process_begin"):
                    r.process_begin(executable, title, handle)

        else:
            # Grammar's context doesn't match, deactivate active rules.
            if self._in_context:
                self._in_context = False
                self.exit_context()
            [r.deactivate() for r in self._rules if r.active]

        self._log_begin.debug("Grammar %s:     active rules: %s.",
                              self._name,
                              [r.name for r in self._rules if r.active])

    def enter_context(self):
        """
            Enter context callback.

            This method is called when a phrase-start has been
            detected.  It is only called if this grammar's
            context previously did not match but now does
            match positively.

        """

    def exit_context(self):
        """
            Exit context callback.

            This method is called when a phrase-start has been
            detected.  It is only called if this grammar's
            context previously did match but now doesn't
            match positively anymore.

        """

    def _process_begin(self, executable, title, handle):
        """
            Start of phrase callback.

            *This usually is the method which should be overridden
            to give derived grammar classes custom behavior.*

            This method is called when the speech recognition
            engine detects that the user has begun to speak a
            phrase. This method is called by the
            ``Grammar.process_begin`` method only if this
            grammar's context matches positively.

            Arguments:
             - *executable* -- the full path to the module whose
               window is currently in the foreground.
             - *title* -- window title of the foreground window.
             - *handle* -- window handle to the foreground window.

        """



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Compound element classes
============================================================================

The following special *element* classes exist as convenient ways of
constructing basic element types from string specifications:

 * :class:`Compound` --
   a special element which parses a string spec to create a hierarchy of
   basic elements.

 * :class:`Choice` --
   a special element taking a :code:`choice` dictionary argument,
   interpreting keys as :class:`Compound` string specifications and values
   for what to return when compound specs are successfully decoded during
   the recognition process.

   The :code:`choice` argument may also be a list or tuple of strings, in
   which case the strings are also interpreted as :class:`Compound` strings
   specifications.  However, the values returned when compound specs are
   successfully decoded during the recognition process are the recognized
   words.  **Note**: these values will be matching part(s) of the compound
   specs.

"""

# pylint: disable=W0223
# Suppress abstract method warnings.

import locale
import logging
import re

from six import string_types, binary_type

from dragonfly.grammar.elements_basic import Alternative, ElementBase
from dragonfly.parsing.parse import spec_parser, CompoundTransformer, ParseError

#---------------------------------------------------------------------------
# The Compound class.

class Compound(Alternative):
    """
        Element which parses a string spec to create a hierarchy of basic
        elements.

        Constructor arguments:
         - *spec* (*str*) -- compound specification
         - *extras* (sequence) -- extras elements referenced from the
           compound spec
         - *actions* (*dict*) -- this argument is currently unused
         - *name* (*str*) -- the name of this element
         - *value* (*object*, default: *None*) --
           value to be returned when this element is successfully decoded
           If *None*, then the value(s) of child nodes are used instead
         - *value_func* (*callable*, default: *None*) --
           function to be called for the node value when this element is
           successfully decoded. If *None*, then the value(s) of child nodes
           are used. This argument takes precedence over the *value*
           argument if it is present
         - *elements* (sequence) -- same as the extras argument
         - *default* (default: *None*) -- the default value of this element

        Example:

        .. code:: python

           # Define a command to type the sum of two spoken integers between
           # 1 and 50 using a Compound element.
           mapping = {
               "type <XAndY>": Text("%(XAndY)d"),
           }
           extras = [
               Compound(
                   # Pass the compound spec and element name.
                   spec="<x> and <y>",
                   name="XAndY",

                   # Pass the internal IntegerRef extras.
                   extras=[IntegerRef("x", 1, 50), IntegerRef("y", 1, 50)],

                   # Pass a value function that adds the two spoken integers
                   # together.
                   value_func=lambda node, extras: extras["x"] + extras["y"])
           ]

    """

    _log = logging.getLogger("compound.parse")
    _parser = spec_parser

    def __init__(self, spec, extras=None, actions=None, name=None,
                 value=None, value_func=None, elements=None, default=None):
        # pylint: disable=too-many-arguments,too-many-branches
        if isinstance(spec, binary_type):
            spec = spec.decode(locale.getpreferredencoding())

        self._spec = spec
        self._value = value
        self._value_func = value_func

        if extras   is None:   extras   = {}
        if actions  is None:   actions  = {}
        if elements is None:   elements = {}

        # Convert extras argument from sequence to mapping.
        if isinstance(extras, (tuple, list)):
            mapping = {}
            for element in extras:
                if not isinstance(element, ElementBase):
                    self._log.error("Invalid extras item: %s", element)
                    raise TypeError("Invalid extras item: %s" % element)
                if not element.name:
                    self._log.error("Extras item does not have a name: %s", element)
                    raise TypeError("Extras item does not have a name: %s" % element)
                if element.name in mapping:
                    self._log.warning("Multiple extras items with the same name: %s", element)
                mapping[element.name] = element
            extras = mapping
        elif not isinstance(extras, dict):
            self._log.error("Invalid extras argument: %s", extras)
            raise TypeError("Invalid extras argument: %s" % extras)

        # Temporary transition code so that both "elements" and "extras"
        #  are supported as keyword arguments.
        if extras and elements:
            extras = dict(extras)
            extras.update(elements)
        elif elements:
            extras = elements
        self._extras = extras

        try:
            tree = self._parser.parse(spec)
        except Exception as e:
            self._log.error("Exception raised parsing %r: %s", spec, e)
            raise ParseError("Exception raised parsing %r: %s" % (spec, e))

        try:
            element = CompoundTransformer(self._extras).transform(tree)
        except Exception as e:
            self._log.error("Exception raised transforming %r: %s", spec, e)
            raise ParseError("Exception raised transforming %r: %s" % (spec, e))

        Alternative.__init__(self, (element,), name=name,
                             default=default)

    def __repr__(self):
        arguments = ["%r" % self._spec]
        if self.name:
            arguments.append("name=%r" % self.name)
        arguments = ", ".join(arguments)
        return "%s(%s)" % (self.__class__.__name__, arguments)

    def value(self, node):
        if self._value_func is not None:
            # Prepare *extras* dict for passing to value_func().
            extras = {"_node": node}
            for name, element in self._extras.items():
                extra_node = node.get_child_by_name(name, shallow=True)
                if extra_node:
                    extras[name] = extra_node.value()
                elif element.has_default():
                    extras[name] = element.default
            try:
                value = self._value_func(node, extras)
            except Exception as e:
                self._log.warning("Exception from value_func: %s", e)
                raise
            return value
        elif self._value is not None:
            return self._value
        else:
            return Alternative.value(self, node)


#---------------------------------------------------------------------------
# The Choice class which maps multiple Compound instances to values.

class Choice(Alternative):
    """
        Element allowing a dictionary of phrases (compound specs) to be
        recognized to be mapped to objects to be used in an action.

        A list or tuple of phrases to be recognized may also be used.  In
        this case the strings are also interpreted as :class:`Compound`
        string specifications.  However, the values returned when compound
        specs are successfully decoded during the recognition process are
        the recognized words.  **Note**: these values will be matching
        part(s) of the compound specs.

        Constructor arguments:
            - *name* (*str*) -- the name of this element
            - *choices* (*dict*, *list* or *tuple*) -- dictionary mapping
              recognized phrases to returned values **or** a list/tuple of
              recognized phrases
            - *extras* (*list*, default: *None*) -- a list of included
              extras
            - *default* (default: *None*) -- the default value of this
              element

        Example using a dictionary:

        .. code:: python

            # Tab switching command, e.g. 'third tab'.
            mapping = {
                "<nth> tab": Key("c-%(nth)s"),
            }
            extras = [
                Choice("nth", {
                    "first"         : "1",
                    "second"        : "2",
                    "third"         : "3",
                    "fourth"        : "4",
                    "fifth"         : "5",
                    "sixth"         : "6",
                    "seventh"       : "7",
                    "eighth"        : "8",
                    "(last | ninth)": "9",
                    "next"          : "pgdown",
                    "previous"      : "pgup",
                }),
            ]

        Example using a list:

        .. code:: python

            # Command for recognizing and typing nth words, e.g.
            # 'type third'.
            mapping = {
                "type <nth>": Text("%(nth)s"),
            }
            extras = [
                Choice("nth", [
                    "first",
                    "second",
                    "third",
                    "fourth",
                    "fifth",
                    "sixth",
                    "seventh",
                    "eighth",

                    # Note that the decoded value for a compound spec like
                    #  this one, when used in a list/tuple of choices,
                    #  rather than a dictionary, is the recognized word:
                    #  "last" or "ninth".
                    "(last | ninth)",

                    "next",
                    "previous",
                ]),
            ]
    """
    def __init__(self, name, choices, extras=None, default=None):
        # Argument type checking.
        assert isinstance(name, string_types) or name is None
        assert isinstance(choices, (dict, list, tuple))
        choices_is_sequence = isinstance(choices, (list, tuple))
        if choices_is_sequence:
            choices = {k: None for k in choices}

        for k, v in choices.items():
            assert isinstance(k, string_types)

        # Construct children from the given choice keys and values.
        self._choices = choices
        self._extras = extras
        children = []
        for k, v in choices.items():
            child = Compound(spec=k, value=v, extras=extras)
            children.append(child)

        # Initialize super class.
        Alternative.__init__(self, children=children,
                             name=name, default=default)



"""
Command-module loader for Kaldi.

This script is based on 'dfly-loader-wsr.py' written by Christo Butcher and
has been adapted to work with the Kaldi engine instead.

This script can be used to look for Dragonfly command-modules for use with
the Kaldi engine. It scans the directory it's in and loads any ``_*.py`` it
finds.
"""


# TODO Have a simple GUI for pausing, resuming, cancelling and stopping
# recognition, etc

from __future__ import print_function

import logging
import os.path
import sys

import six

from dragonfly import get_engine
from dragonfly import Grammar, MappingRule, Function, Dictation, FuncContext
from dragonfly.loader import CommandModuleDirectory
from dragonfly.log import setup_log


# --------------------------------------------------------------------------
# Set up basic logging.

if False:
    # Debugging logging for reporting trouble
    logging.basicConfig(level=10)
    logging.getLogger('grammar.decode').setLevel(20)
    logging.getLogger('grammar.begin').setLevel(20)
    logging.getLogger('compound').setLevel(20)
    logging.getLogger('kaldi.compiler').setLevel(10)
else:
    setup_log()


# --------------------------------------------------------------------------
# User notification / rudimentary UI. MODIFY AS DESIRED

# For message in ('sleep', 'wake')
def notify(message):
    if message == 'sleep':
        print("Sleeping...")
        # get_engine().speak("Sleeping")
    elif message == 'wake':
        print("Awake...")
        # get_engine().speak("Awake")


# --------------------------------------------------------------------------
# Sleep/wake grammar. (This can be unused or removed if you don't want it.)

sleeping = False

def load_sleep_wake_grammar(initial_awake):
    sleep_grammar = Grammar("sleep")

    def sleep(force=False):
        global sleeping
        if not sleeping or force:
            sleeping = True
            sleep_grammar.set_exclusiveness(True)
        notify('sleep')

    def wake(force=False):
        global sleeping
        if sleeping or force:
            sleeping = False
            sleep_grammar.set_exclusiveness(False)
        notify('wake')

    class SleepRule(MappingRule):
        mapping = {
            "start listening":  Function(wake) + Function(lambda: get_engine().start_saving_adaptation_state()),
            "stop listening":   Function(lambda: get_engine().stop_saving_adaptation_state()) + Function(sleep),
            "halt listening":   Function(lambda: get_engine().stop_saving_adaptation_state()) + Function(sleep),
        }
    sleep_grammar.add_rule(SleepRule())

    sleep_noise_rule = MappingRule(
        name = "sleep_noise_rule",
        mapping = { "<text>": Function(lambda text: False and print(text)) },
        extras = [ Dictation("text") ],
        context = FuncContext(lambda: sleeping),
    )
    sleep_grammar.add_rule(sleep_noise_rule)

    sleep_grammar.load()

    if initial_awake:
        wake(force=True)
    else:
        sleep(force=True)


# --------------------------------------------------------------------------
# Main event driving loop.

def main():
    logging.basicConfig(level=logging.INFO)

    try:
        path = os.path.dirname(__file__)
    except NameError:
        # The "__file__" name is not always available, for example
        # when this module is run from PythonWin.  In this case we
        # simply use the current working directory.
        path = os.getcwd()
        __file__ = os.path.join(path, "kaldi_module_loader_plus.py")

    # Set any configuration options here as keyword arguments.
    # See Kaldi engine documentation for all available options and more info.
    engine = get_engine('kaldi',
        # model_dir='kaldi_model',  # default model directory
        # vad_aggressiveness=3,  # default aggressiveness of VAD
        # vad_padding_start_ms=150,  # default ms of required silence before VAD
        # vad_padding_end_ms=150,  # default ms of required silence after VAD
        # vad_complex_padding_end_ms=500,  # default ms of required silence after VAD for complex utterances
        # input_device_index=None,  # set to an int to choose a non-default microphone
        # lazy_compilation=True,  # set to True to parallelize & speed up loading
        # retain_dir=None,  # set to a writable directory path to retain recognition metadata and/or audio data
        # retain_audio=None,  # set to True to retain speech data wave files in the retain_dir (if set)
    )

    # Call connect() now that the engine configuration is set.
    engine.connect()

    # Load grammars.
    load_sleep_wake_grammar(True)
    directory = CommandModuleDirectory(path, excludes=[__file__])
    directory.load()

    # Define recognition callback functions.
    def on_begin():
        print("Speech start detected.")

    def on_recognition(words):
        message = u"Recognized: %s" % u" ".join(words)

        # This only seems to be an issue with Python 2.7 on Windows.
        if six.PY2:
            encoding = sys.stdout.encoding or "ascii"
            message = message.encode(encoding, errors='replace')
        print(message)

    def on_failure():
        print("Sorry, what was that?")

    # Start the engine's main recognition loop
    engine.prepare_for_recognition()
    try:
        print("Listening...")
        engine.do_recognition(on_begin, on_recognition, on_failure)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dragonfly's logging infrastructure is defined in the ``dragonfly.log``
module. It defines sane defaults for the various loggers used in the library
as well as functions for setting up logging and tracing.

Adjusting logger levels
----------------------------------------------------------------------------

Dragonfly's logger levels can be adjusted much like Python logger levels:

..  code::

    import logging
    logging.getLogger("engine").setLevel(logging.DEBUG)

The one caveat is that this must be done *after* the :meth:`setup_log`
function is called, otherwise the levels you set will be overridden.
By default, the function is only called near the top of the module loader
scripts (e.g. *dragonfly/examples/dfly-loader-wsr.py*), not within dragonfly
itself.

It is not necessary to call :meth:`setup_log` at all. Standard Python
logging functions such as :meth:`basicConfig` can be used at the top of
module loader scripts instead.

If you are not using dragonfly with a module loader, you will need to set up
a logging handler to avoid messages like tho following::

    No handlers could be found for logger "typeables"


Functions
----------------------------------------------------------------------------

"""

import sys
import os.path
import logging


# --------------------------------------------------------------------------
# Sane defaults for logger names and associated levels.

_debug     = logging.DEBUG
_info      = logging.INFO
_warning   = logging.WARNING
_error     = logging.ERROR
_critical  = logging.CRITICAL
default_levels = {
                  "":                     (_warning, _warning),
                  "engine":               (_info, _info),
                  "engine.compiler":      (_warning, _info),
                  "engine.timer":         (_warning, _info),
                  "grammar":              (_warning, _critical),
                  "grammar.load":         (_warning, _info),
                  "grammar.begin":        (_info, _info),
                  "grammar.results":      (_warning, _warning),
                  "grammar.decode":       (_warning, _info),
                  "grammar.eval":         (_warning, _warning),
                  "grammar.process":      (_warning, _warning),
                  "lang":                 (_warning, _info),
                  "compound.parse":       (_warning, _info),
                  "dictation.formatter":  (_warning, _warning),
                  "dictation.word_parser":  (_warning, _warning),
                  "dictation.word_parser_factory":  (_warning, _warning),
                  "action":               (_warning, _warning),
                  "action.init":          (_warning, _warning),
                  "action.exec":          (_warning, _warning),
                  "context":              (_warning, _info),
                  "context.match":        (_warning, _info),
                  "rpc.server":           (_warning, _info),
                  "rpc.methods":          (_warning, _info),
                  "werkzeug":             (_warning, _warning),
                  "jsonrpc.manager":      (_critical, _critical),
                  "rule":                 (_warning, _info),
                  "clipboard":            (_warning, _info),
                  "config":               (_warning, _info),
                  "module":               (_info, _info),
                  "monitor.init":         (_warning, _info),
                  "dfly.test":            (_debug, _debug),
                  "accessibility":        (_info, _info),
                  "keyboard":             (_warning, _warning),
                  "typeables":            (_warning, _warning),
                 }


# --------------------------------------------------------------------------
# Logging filter class which filters out messages of a given name below
#  a given level.

class NameLevelFilter(logging.Filter):
    """"""
    def __init__(self, name, level):
        self.name = name
        self.level = level
        super(NameLevelFilter, self).__init__(name)

    def filter(self, record):
        if record.name == self.name:
            if record.levelno >= self.level:
                return True
            else:
                return False
        else:
            return True


# --------------------------------------------------------------------------

class DispatchingHandler(logging.Handler):
    """"""

    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self.handler_filter_pairs = []

    def filter(self, record):
        return True

    def add_handler_filter_pair(self, handler, filter):
        self.handler_filter_pairs.append((handler, filter))

    def emit(self, record):
        # print "dispatching", self, self.handler_filter_pairs, record
        # import traceback; print traceback.extract_stack()
        # import traceback; traceback.print_stack()
        for handler, filter in self.handler_filter_pairs:
            if filter.filter(record):
                handler.handle(record)


# --------------------------------------------------------------------------

def _setup_stderr_handler():
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(name)s (%(levelname)s): %(message)s")
    stderr_handler.setFormatter(formatter)
    return stderr_handler

_file_handler = None


def _setup_file_handler():
    global _file_handler
#    import traceback; traceback.print_stack()
    if not _file_handler:
        # Use ~/.dragonfly.log as the file for Dragonfly's log messages.
        home_path = os.path.expanduser("~")
        log_file_path = os.path.join(home_path, ".dragonfly.log")
        _file_handler = logging.FileHandler(log_file_path)
        formatter = logging.Formatter("%(asctime)s %(name)s (%(levelname)s):"
                                      " %(message)s")
        _file_handler.setFormatter(formatter)
    return _file_handler


# --------------------------------------------------------------------------

_stderr_handler = None
_dispatching_handlers = {}
_stderr_filters = {}
_file_filters = {}


def setup_log(use_stderr=True, use_file=True, use_stdout=False):
    """
    Setup Dragonfly's logging infrastructure with sane defaults.

    :param use_stderr: whether to output log messages to stderr
      (default: True).
    :param use_file: whether to output log messages to the
      *~/.dragonfly.log* log file (default: True).
    :param use_stdout: this parameter does nothing andhas been left in for
      backwards-compatibility (default: False).
    :type use_stderr: bool
    :type use_file: bool
    :type use_stdout: bool

    """
    global _dispatching_handlers
    global _stderr_handler, _file_handler
    global _stderr_filters, _file_filters

    # Remove any previously created dispatching handlers.
    for name, handler in _dispatching_handlers.items():
        logger = logging.getLogger(name)
        logger.removeHandler(handler)

    # Setup default handlers.
    if use_stderr:
        _stderr_handler = _setup_stderr_handler()
    if use_file:
        _file_handler = _setup_file_handler()

    # Create and register default filters.
    for name, levels in default_levels.items():
        stderr_level, file_level = levels
        handler = DispatchingHandler()
        _dispatching_handlers[name] = handler
        if use_stderr:
            stderr_filter = NameLevelFilter(name, stderr_level)
            handler.add_handler_filter_pair(_stderr_handler, stderr_filter)
            _stderr_filters[name] = stderr_filter
        if use_file:
            file_filter = NameLevelFilter(name, file_level)
            handler.add_handler_filter_pair(_file_handler, file_filter)
            _file_filters[name] = file_filter
        logger = logging.getLogger(name)
        logger.addHandler(handler)
        logger.setLevel(min(stderr_level, file_level))
        logger.propagate = False


# --------------------------------------------------------------------------
# Function for setting up call tracing for low-level debugging.

def setup_tracing(output, limit=None):
    """
    Setup call tracing for low-level debugging.

    :param output: the file to write tracing messages to.
    :type output: file
    :param limit: the recursive depth limit for tracing (default: None).
    :type limit: int|None
    """
    from pkg_resources import resource_filename
    library_prefix = os.path.dirname(resource_filename(__name__, "setup.py"))
    print("prefix:", library_prefix)
    exclude_filenames = ("parser.py",)

    def _tracing_callback(frame, event, arg):
        # Retrieve current function name, line number, etc.
        code_object = frame.f_code
        function_name = code_object.co_name
        line_number = frame.f_lineno
        filename = code_object.co_filename
        if not filename.startswith(library_prefix):
            return
        else:
            filename = filename[len(library_prefix)+1:]
        if os.path.basename(filename) in exclude_filenames:
            return

        depth = 0
        if limit is not None:
            # Determine call depth of current frame.
            parent_frame = frame
            while parent_frame:
                parent_frame = parent_frame.f_back
                depth += 1
            del parent_frame
            if depth > limit:
                return

        # Write message to output.
        indented_function_name = ("  " * depth) + function_name
        output.write("%2d %-40s %5s %-40s\n" % (depth, indented_function_name,
                                                line_number, filename))
        output.flush()

    sys.settrace(_tracing_callback)



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
SR back-end for DNS and Natlink
============================================================================

Detecting sleep mode
----------------------------------------------------------------------------

 - http://blogs.msdn.com/b/tsfaware/archive/2010/03/22/detecting-sleep-mode-in-sapi.aspx

"""

import os
import os.path
import pywintypes
import sys
import time
from datetime import datetime
from locale import getpreferredencoding
from threading import Thread, Event

from six import text_type, binary_type, string_types, PY2

from ..base        import (EngineBase, EngineError, MimicFailure,
                           GrammarWrapperBase)
from .compiler     import NatlinkCompiler
from .dictation    import NatlinkDictationContainer
from .recobs       import NatlinkRecObsManager
from .timer        import NatlinkTimerManager
import dragonfly.grammar.elements as elements_
import dragonfly.grammar.state as state_


# ---------------------------------------------------------------------------

def map_word(word, encoding=getpreferredencoding(do_setlocale=False)):
    """
    Wraps output from Dragon.

    This wrapper ensures text output from the engine is Unicode. It assumes the
    encoding of byte streams is the current locale's preferred encoding by default.
    """
    if isinstance(word, text_type):
        return word
    elif isinstance(word, binary_type):
        return word.decode(encoding)
    return word


class TimerThread(Thread):
    """"""
    def __init__(self, engine):
        Thread.__init__(self)
        self._stop_event = Event()
        self.daemon = True
        self._timer = None
        self._engine = engine

    def start(self):
        if self._timer is None:
            def timer_function():
                # Let the thread run for a bit. This will yield control to
                # other threads.
                if self.is_alive():
                    self.join(0.0025)

            self._timer = self._engine.create_timer(timer_function, 0.025)

        Thread.start(self)

    def _stop_timer(self):
        if self._timer:
            self._timer.stop()
            self._timer = None

    def stop(self):
        self._stop_event.set()
        self._stop_timer()

    def run(self):
        while not self._stop_event.is_set():
            time.sleep(1)
        self._stop_timer()


class NatlinkEngine(EngineBase):
    """ Speech recognition engine back-end for Natlink and DNS. """

    _name = "natlink"
    DictationContainer = NatlinkDictationContainer

    #-----------------------------------------------------------------------

    def __init__(self, retain_dir=None):
        """
        :param retain_dir: directory to save audio data:
          A ``.wav`` file for each utterance, and ``retain.tsv`` file
          with each row listing (wav filename, wav length in seconds,
          grammar name, rule name, recognized text) as tab separated
          values.

          If this parameter is used in a module loaded by
          ``natlinkmain``, then the directory will be relative to the
          Natlink user directory (e.g. ``MacroSystem``).
        :type retain_dir: str|None
        """
        EngineBase.__init__(self)

        self.natlink = None
        try:
            import natlink
        except ImportError:
            self._log.error("%s: failed to import natlink module." % self)
            raise EngineError("Requested engine 'natlink' is not "
                              "available: Natlink is not installed.")
        self.natlink = natlink

        self._grammar_count = 0
        self._recognition_observer_manager = NatlinkRecObsManager(self)
        self._timer_manager = NatlinkTimerManager(0.02, self)
        self._timer_thread = None
        self._retain_dir = None
        try:
            self.set_retain_directory(retain_dir)
        except EngineError as err:
            self._retain_dir = None
            self._log.error(err)

    def apply_threading_fix(self):
        """
        Start a thread and engine timer internally to allow Python threads
        to work properly while connected to natlink. The fix is only applied
        once, successive calls have no effect.

        This method is called automatically when :meth:`connect` is called
        or when a grammar is loaded for the first time.
        """
        # Start a thread and engine timer to allow Python threads to work
        # properly while connected to Natlink.
        # Only start the thread if one isn't already active.
        if self._timer_thread is None:
            self._timer_thread = TimerThread(self)
            self._timer_thread.start()

    def connect(self):
        """ Connect to natlink with Python threading support enabled. """
        self.natlink.natConnect(True)
        self.apply_threading_fix()

    def disconnect(self):
        """ Disconnect from natlink. """
        # Unload all grammars from the engine so that Dragon doesn't keep
        # recognizing them.
        for grammar in self.grammars:
            grammar.unload()

        # Close the the waitForSpeech() dialog box if it is active for this
        # process.
        from dragonfly import Window
        target_title = "Natlink / Python Subsystem"
        for window in Window.get_matching_windows(title=target_title):
            if window.is_visible and window.pid == os.getpid():
                try:
                    window.close()
                except pywintypes.error:
                    pass
                break

        # Stop the special timer thread if it is running.
        if self._timer_thread:
            self._timer_thread.stop()
            self._timer_thread = None

        # Finally disconnect from natlink.
        self.natlink.natDisconnect()

    # -----------------------------------------------------------------------
    # Methods for working with grammars.

    def _load_grammar(self, grammar):
        """ Load the given *grammar* into natlink. """
        self._log.debug("Engine %s: loading grammar %s."
                        % (self, grammar.name))

        grammar_object = self.natlink.GramObj()
        wrapper = GrammarWrapper(grammar, grammar_object, self,
                                 self._recognition_observer_manager)
        grammar_object.setBeginCallback(wrapper.begin_callback)
        grammar_object.setResultsCallback(wrapper.results_callback)
        grammar_object.setHypothesisCallback(None)

        c = NatlinkCompiler()
        (compiled_grammar, rule_names) = c.compile_grammar(grammar)
        grammar._rule_names = rule_names

        all_results = (hasattr(grammar, "process_recognition_other")
                       or hasattr(grammar, "process_recognition_failure"))
        hypothesis = False

        attempt_connect = False
        try:
            grammar_object.load(compiled_grammar, all_results, hypothesis)
        except self.natlink.NatError as e:
            # If loading failed because we're not connected yet,
            #  attempt to connect to natlink and reload the grammar.
            if (str(e) == "Calling GramObj.load is not allowed before"
                          " calling natConnect"):
                attempt_connect = True
            else:
                self._log.exception("Failed to load grammar %s: %s."
                                    % (grammar, e))
                raise EngineError("Failed to load grammar %s: %s."
                                  % (grammar, e))
        if attempt_connect:
            self.connect()
            try:
                grammar_object.load(compiled_grammar, all_results, hypothesis)
            except self.natlink.NatError as e:
                self._log.exception("Failed to load grammar %s: %s."
                                    % (grammar, e))
                raise EngineError("Failed to load grammar %s: %s."
                                  % (grammar, e))

        # Apply the threading fix if it hasn't been applied yet.
        self.apply_threading_fix()

        # Return the grammar wrapper.
        return wrapper

    def _unload_grammar(self, grammar, wrapper):
        """ Unload the given *grammar* from natlink. """
        try:
            grammar_object = wrapper.grammar_object
            grammar_object.unload()
            grammar_object.setBeginCallback(None)
            grammar_object.setResultsCallback(None)
            grammar_object.setHypothesisCallback(None)
        except self.natlink.NatError as e:
            self._log.exception("Failed to unload grammar %s: %s."
                                % (grammar, e))

    def set_exclusiveness(self, grammar, exclusive):
        try:
            grammar_object = self._get_grammar_wrapper(grammar).grammar_object
            grammar_object.setExclusive(exclusive)
        except self.natlink.NatError as e:
            self._log.exception("Engine %s: failed set exclusiveness: %s."
                                % (self, e))

    def activate_grammar(self, grammar):
        self._log.debug("Activating grammar %s." % grammar.name)
        pass

    def deactivate_grammar(self, grammar):
        self._log.debug("Deactivating grammar %s." % grammar.name)
        pass

    def activate_rule(self, rule, grammar):
        self._log.debug("Activating rule %s in grammar %s." % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        grammar_object = wrapper.grammar_object
        grammar_object.activate(rule.name, 0)

    def deactivate_rule(self, rule, grammar):
        self._log.debug("Deactivating rule %s in grammar %s." % (rule.name, grammar.name))
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        grammar_object = wrapper.grammar_object
        grammar_object.deactivate(rule.name)

    def update_list(self, lst, grammar):
        wrapper = self._get_grammar_wrapper(grammar)
        if not wrapper:
            return
        grammar_object = wrapper.grammar_object

        # First empty then populate the list.  Use the local variables
        #  n and f as an optimization.
        n = lst.name
        f = grammar_object.appendList
        grammar_object.emptyList(n)
        [f(n, word) for word in lst.get_list_items()]

        # Clear grammar wrapper word sets so they get recalculated.
        wrapper.rule_words_map.clear()

    #-----------------------------------------------------------------------
    # Miscellaneous methods.

    def _do_recognition(self):
        self.natlink.waitForSpeech()

    def mimic(self, words):
        """
        Mimic a recognition of the given *words*.

        .. note:: This method has a few quirks to be aware of:

           #. Mimic is not limited to one element per word as seen with
              proper nouns from DNS. For example, "Buffalo Bills" can be
              passed as one word.
           #. Mimic can handle by the extra formatting by DNS built-in
              commands.
           #. Mimic is case sensitive.

        """
        if isinstance(words, string_types):
            words = words.split()

        try:
            prepared_words = []
            if PY2:
                encoding = getpreferredencoding()
                for word in words:
                    if isinstance(word, text_type):
                        word = word.encode(encoding)
                    prepared_words.append(word)
            else:
                for word in words:
                    prepared_words.append(word)
            if len(prepared_words) == 0:
                raise TypeError("empty list or string")
        except Exception as e:
            raise MimicFailure("Invalid mimic input %r: %s."
                               % (words, e))
        try:
            self.natlink.recognitionMimic(prepared_words)
        except self.natlink.MimicFailed:
            raise MimicFailure("No matching rule found for words %r."
                               % (prepared_words,))

    def speak(self, text):
        """ Speak the given *text* using text-to-speech. """
        # Store the current mic state.
        mic_state = self.natlink.getMicState()

        # Say the text.
        self.natlink.execScript('TTSPlayString "%s"' % text)

        # Restore the previous mic state if necessary.
        # This is to have consistent behaviour for each version of Dragon.
        if mic_state != self.natlink.getMicState():
            self.natlink.setMicState(mic_state)

    def _get_language(self):
        # Get a Windows language identifier from Dragon.
        import win32com.client
        app = win32com.client.Dispatch("Dragon.DgnEngineControl")
        language = app.SpeakerLanguage("")

        # Lookup and return the language tag.
        return self._get_language_tag(language)

    def _has_quoted_words_support(self):
        return True

    def set_retain_directory(self, retain_dir):
        """
        Set the directory where audio data is saved.

        Retaining audio data may be useful for acoustic model training. This
        is disabled by default.

        If a relative path is used and the code is running via natspeak.exe,
        then the path will be made relative to the Natlink user directory or
        base directory (e.g. ``MacroSystem``).

        :param retain_dir: retain directory path
        :type retain_dir: string|None
        """
        is_string = isinstance(retain_dir, string_types)
        if not (retain_dir is None or is_string):
            raise EngineError("Invalid retain_dir: %r" % retain_dir)

        if is_string:
            # Handle relative paths by using the Natlink user directory or
            # base directory. Only do this if running via natspeak.exe.
            try:
                import natlinkstatus
            except ImportError:
                natlinkstatus = None
            running_via_natspeak = (
                sys.executable.endswith("natspeak.exe") and
                natlinkstatus is not None
            )
            if not os.path.isabs(retain_dir) and running_via_natspeak:
                status = natlinkstatus.NatlinkStatus()
                user_dir = status.getUserDirectory()
                retain_dir = os.path.join(
                    # Use the base dir if user dir isn't set.
                    user_dir if user_dir else status.BaseDirectory,
                    retain_dir
                )

        # Set the retain directory.
        self._retain_dir = retain_dir

#---------------------------------------------------------------------------


class GrammarWrapper(GrammarWrapperBase):

    def __init__(self, grammar, grammar_object, engine, recobs_manager):
        GrammarWrapperBase.__init__(self, grammar, engine, recobs_manager)
        self.grammar_object = grammar_object
        self.rule_words_map = {}

    def get_rule_words(self, rule):
        # Return a set containing any words used in this rule or in any
        #  referenced rules or lists. Store the set for each rule as an
        #  optimization.
        if rule.name in self.rule_words_map:
            return self.rule_words_map[rule.name]

        words = set()
        for element in self.grammar._get_element_list(rule):
            if isinstance(element, elements_.Literal):
                # Only get the required first word.
                literal_words = element.words_ext
                if literal_words:
                    words.add(literal_words[0])
            elif isinstance(element, elements_.RuleRef):
                words.update(self.get_rule_words(element.rule))
            elif isinstance(element, elements_.ListRef):
                for string in element.list.get_list_items():
                    # Only get the required first word.
                    list_item_words = string.split()
                    if list_item_words:
                        words.add(list_item_words[0])

        self.rule_words_map[rule.name] = words
        return words

    def begin_callback(self, module_info):
        executable, title, handle = tuple(map_word(word)
                                          for word in module_info)
        self.grammar.process_begin(executable, title, handle)

    def _process_rules(self, words, words_rules, results,
                       manual_rule_ids):
        # Iterates through this grammar's rules, attempting
        #  to decode each.  If successful, call that rule's
        #  method for processing the recognition and return.
        s = state_.State(words_rules, self.grammar._rule_names,
                         self.engine)
        for r in self.grammar._rules:
            if not (r.active and r.exported): continue

            # Set dictation words manually if DNS didn't report a difference
            #  between command and dictation words. A word is set as
            #  dictation if it isn't a reported DNS dictation word and isn't
            #  a word in the current top-level rule or any referenced rules.
            if manual_rule_ids:
                rule_words = self.get_rule_words(r)
                words_rules2 = tuple(
                    (w, 1000000) if r < 1000000 and w not in rule_words
                    else (w, r)
                    for w, r in words_rules
                )
                s = state_.State(words_rules2, self.grammar._rule_names,
                                 self.engine)
            s.initialize_decoding()
            for result in r.decode(s):
                if s.finished():
                    self._retain_audio(words, results, r.name)
                    root = s.build_parse_tree()

                    # Notify observers using the manager *before*
                    # processing.
                    notify_args = (words, r, root, results)
                    self.recobs_manager.notify_recognition(*notify_args)

                    r.process_recognition(root)

                    # Notify observers using the manager *after*
                    # processing.
                    self.recobs_manager.notify_post_recognition(
                        *notify_args
                    )
                    return True

        return False

    def results_callback(self, words, results):
        NatlinkEngine._log.debug("Grammar %s: received recognition %r."
                                 % (self.grammar._name, words))

        if words == "other":
            func = getattr(self.grammar, "process_recognition_other", None)
            self._process_grammar_callback(
                func, words=tuple(map_word(w) for w in results.getWords(0)),
                results=results
            )
            return
        elif words == "reject":
            func = getattr(self.grammar, "process_recognition_failure",
                           None)
            self._process_grammar_callback(func, results=results)
            return

        # If the words argument was not "other" or "reject", then
        #  it is a sequence of (word, rule_id) 2-tuples.  Convert this
        #  into a tuple of unicode objects.
        words_rules = tuple((map_word(w), r) for w, r in words)
        words = tuple(w for w, r in words_rules)

        # Call the grammar's general process_recognition method, if present.
        func = getattr(self.grammar, "process_recognition", None)
        if func:
            if not self._process_grammar_callback(func, words=words,
                                                  results=results):
                # Return early if the method didn't return True or equiv.
                return

        # Attempt to decode each grammar rule and process the recognition if
        #  successful.
        if self._process_rules(words, words_rules, results, False):
            return

        # Try again. This time try to set words as dictation words where
        #  appropriate.
        if self._process_rules(words, words_rules, results, True):
            return

        # Failed to decode recognition.
        NatlinkEngine._log.warning("Grammar %s: failed to decode"
                                   " recognition %r."
                                   % (self.grammar._name, words))

    def _retain_audio(self, words, results, rule_name):
        # Only write audio data and metadata if the directory exists.
        retain_dir = self.engine._retain_dir
        if retain_dir and not os.path.isdir(retain_dir):
            self.engine._log.warning(
                "Audio was not retained because '%s' was not a "
                "directory" % retain_dir
            )
        elif retain_dir:
            try:
                audio = results.getWave()
                # Make sure we have audio data
                if len(audio) > 0:
                    # Write audio data.
                    now = datetime.now()
                    filename = ("retain_%s.wav"
                                % now.strftime("%Y-%m-%d_%H-%M-%S_%f"))
                    wav_path = os.path.join(retain_dir, filename)
                    with open(wav_path, "wb") as f:
                        f.write(audio)

                    # Write metadata, assuming 11025Hz 16bit mono audio
                    text = ' '.join(words)
                    audio_length = float(len(audio) / 2) / 11025
                    tsv_path = os.path.join(retain_dir, "retain.tsv")
                    with open(tsv_path, "a") as tsv_file:
                        tsv_file.write('\t'.join([
                            filename, text_type(audio_length),
                            self.grammar.name, rule_name, text
                        ]) + '\n')
            except:
                self.engine._log.exception("Exception retaining audio")



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Miscellaneous classes and functions used by the engine, but placed in
another file to reduce the number of lines in engine.py.

"""

import os

from sphinxwrapper import DefaultConfig

from dragonfly import RecognitionObserver


def get_decoder_config_object():
    """ Get the default Sphinx decoder object for the engine. """
    decoder_config = DefaultConfig()

    # Silence the decoder output by default.
    decoder_config.set_string("-logfn", os.devnull)

    # Set voice activity detection configuration options for the decoder.
    # These values can be changed if noise in the background triggers speech
    # start and/or false recognitions (e.g. of short words) frequently.
    # Descriptions for VAD config options were retrieved from:
    # https://cmusphinx.github.io/doc/sphinxbase/fe_8h_source.html

    # Number of silence frames to keep after from speech to silence.
    decoder_config.set_int("-vad_postspeech", 30)

    # Number of speech frames to keep before silence to speech.
    decoder_config.set_int("-vad_prespeech", 20)

    # Number of speech frames to trigger vad from silence to speech.
    decoder_config.set_int("-vad_startspeech", 10)

    # Threshold for decision between noise and silence frames. Log-ratio
    # between signal level and noise level.
    decoder_config.set_float("-vad_threshold", 3.0)
    return decoder_config


class EngineConfig(object):
    """ Default engine configuration. """

    # Configuration for the Pocket Sphinx decoder.
    DECODER_CONFIG = get_decoder_config_object()

    # User language for the engine to use.
    LANGUAGE = "en"

    # Configuration for wake/sleep phrases
    START_ASLEEP = True
    WAKE_PHRASE = "wake up"
    WAKE_PHRASE_THRESHOLD = 1e-20
    SLEEP_PHRASE = "go to sleep"
    SLEEP_PHRASE_THRESHOLD = 1e-40

    # Configuration for acoustic model training.
    TRAINING_DATA_DIR = ""
    TRANSCRIPT_NAME = "training"
    START_TRAINING_PHRASE = "start training session"
    START_TRAINING_PHRASE_THRESHOLD = 1e-48
    END_TRAINING_PHRASE = "end training session"
    END_TRAINING_PHRASE_THRESHOLD = 1e-45

    # Audio input configuration.
    # These values should match the requirements for the acoustic model
    # being used. Default values match the 16 kHz CMU US English models.
    CHANNELS = 1              # one channel (mono)
    SAMPLE_WIDTH = 2          # Sample width of 2 bytes/16 bits
    RATE = 16000              # 16kHz sample rate
    FRAMES_PER_BUFFER = 2048  # frames per audio buffer


class WaveRecognitionObserver(RecognitionObserver):
    """ Observer class used in :meth:`SphinxEngine.process_wave_file`. """
    def __init__(self, engine):
        RecognitionObserver.__init__(self)
        self.words = ""
        self.complete = None
        self.engine = engine

    def on_start(self):
        self.complete = False

    def on_recognition(self, words):
        self.complete = True
        self.words = words

    def on_failure(self):
        self.complete = True

        # Use the default search result on failure.
        hyp = self.engine.default_search_result
        if hyp:
            # print(hyp.prob)
            self.words = hyp.hypstr

    def __enter__(self):
        self.register()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unregister()



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
This file implements base classes for structured integer grammar
elements.

"""

# pylint: disable=redefined-builtin,no-self-use,too-many-branches

import copy

from ...grammar.elements  import (Alternative, Sequence, Optional, RuleRef,
                                  Compound, ListRef, Literal, Impossible)
from ...grammar.list      import List


#---------------------------------------------------------------------------
# Numeric element builder classes.

class IntBuilderBase(object):

    def __init__(self, modifier_function=None, modifier_mode=None):
        if modifier_mode is None:
            modifier_mode = ModifiedPathsCollection.MODE_AUGMENT

        self._modifier_function = modifier_function
        self._modifier_mode = modifier_mode

    def build_element(self, min, max, memo=None):
        if memo is None:
            memo = {}

        # Check if an appropriate element already exists in the memo.
        #  If so, then return it.
        key = (id(self), min, max)
        if key in memo:
            return copy.copy(memo[key])

        # Otherwise, build the element, save it in the memo and return it.
        element = self._build_element(min, max, memo)
        memo[key] = element
        return element

    def _build_element(self, min, max, memo):
        raise NotImplementedError("Call to virtual method _build_element()"
                                  " in base class IntBuilderBase")

    def _build_modified_paths(self, children, memo):
        # pylint: disable=unused-argument
        if len(children) == 0:    return None
        if len(children) == 1:    root = children[0]
        else:                     root = Alternative(children)
        return ModifiedPathsCollection(root, self._modifier_function,
                                       self._modifier_mode)


class MapIntBuilder(IntBuilderBase):

    def __init__(self, mapping):
        self._mapping = mapping
        IntBuilderBase.__init__(self)

    def _build_element(self, min, max, memo):
        mapping_memo = {}
        children = []
        for spec, value in self._mapping.items():
            if min <= value < max:
                if spec in mapping_memo:
                    children.append(mapping_memo[spec])
                else:
                    element = Compound(spec=spec, value=value)
                    children.append(element)
                    mapping_memo[spec] = element
        if len(children) > 1:
            return Alternative(children)
        elif len(children) == 1:
            return children[0]
        else:
            return None


class CollectionIntBuilder(IntBuilderBase):

    def __init__(self, spec, set, modifier_function=None,
                 modifier_mode=None):
        self._spec = spec
        self._set = set
        IntBuilderBase.__init__(self, modifier_function, modifier_mode)

    def _build_element(self, min, max, memo):
        child = self._build_range_set(self._set, min, max, memo)
        if not child:
            return None
        child.name = "element"
        element = Collection(self._spec, child)
        return element

    def _build_range_set(self, set, min, max, memo):
        # Iterate through the set allowing each item to build an element.
        children = [c.build_element(min, max, memo) for c in set]
        children = [c for c in children if c]

        # Build modified path elements, if necessary.
        if self._modifier_function is not None:
            c = self._build_modified_paths(children, memo)
            if c:
                # Don't use other children if in modifier mode REPLACE.
                mode = self._modifier_mode
                if mode == ModifiedPathsCollection.MODE_REPLACE:
                    del children[:]

                # Add the ModifiedPathsCollection element.
                children.append(c)

        # Wrap up results appropriately.
        if not children:
            return None
        if len(children) == 1:
            return children[0]
        else:
            return Alternative(children)


class MagnitudeIntBuilder(IntBuilderBase):

    _empty_list = None

    @classmethod
    def _get_empty_list(cls):
        """ Class method to ensure only one empty list is created. """
        if cls._empty_list is None:
            cls._empty_list = List("_MagnitudeIntBuilder_empty")
        return cls._empty_list

    def __init__(self, factor, spec, multipliers, remainders,
                 modifier_function=None, modifier_mode=None):
        self._factor = factor
        self._spec = spec
        self._multipliers = multipliers
        self._remainders = remainders
        IntBuilderBase.__init__(self, modifier_function, modifier_mode)

    def _build_element(self, min, max, memo):

        # Sanity check.
        if min >= max: return None

        # Calculate ranges of multipliers and remainders.
        first_multiplier = int(min / self._factor)
        last_multiplier  = int((max - 1) / self._factor + 1)
        first_remainder  = min % self._factor
        last_remainder   = max % self._factor
        if last_remainder == 0:
            last_remainder = self._factor

        # Handle special case of only one possible multiplier value.
        if first_multiplier == last_multiplier - 1:
            return self._build_range(first_multiplier, last_multiplier,
                                     first_remainder, last_remainder, memo)

        children = []

        # Build partial range for first multiplier value, if necessary.
        if first_remainder > 0:
            c = self._build_range(first_multiplier, first_multiplier + 1,
                                  first_remainder, self._factor, memo)
            if c: children.append(c)
            first_multiplier += 1

        # Build partial range for last multiplier value, if necessary.
        if last_remainder > 0:
            c = self._build_range(last_multiplier - 1, last_multiplier,
                                  0, last_remainder, memo)
            if c: children.append(c)
            last_multiplier -= 1

        # Build range for multiplier values which have the full
        #  range of remainder values.
        if first_multiplier < last_multiplier:
            c = self._build_range(first_multiplier, last_multiplier,
                                  0, self._factor, memo)
            if c: children.append(c)

        # Build modified path elements, if necessary.
        if self._modifier_function is not None:
            c = self._build_modified_paths(children, memo)
            if c:
                # Don't use other children if in modifier mode REPLACE.
                mode = self._modifier_mode
                if mode == ModifiedPathsCollection.MODE_REPLACE:
                    del children[:]

                # Add the ModifiedPathsCollection element.
                children.append(c)

        # Wrap up result as is appropriate.
        if len(children) == 0:    return None
        elif len(children) == 1:  return children[0]
        else:                     return Alternative(children)

    def _build_range(self, first_multiplier, last_multiplier,
                     first_remainder, last_remainder, memo):

        # Build range for multipliers.
        multipliers = self._build_range_set(self._multipliers,
                                            first_multiplier,
                                            last_multiplier, memo)
        if not multipliers: return None

        # Build range for remainders.
        remainders = self._build_range_set(self._remainders,
                                           first_remainder,
                                           last_remainder, memo)
        if not remainders:
            empty = self._get_empty_list()
            remainders = ListRef("_MagnitudeIntBuilder_empty_ref", empty)

        # Build magnitude element.
        multipliers.name = "multiplier"
        remainders.name = "remainder"
        return Magnitude(self._factor, self._spec, multipliers, remainders)

    def _build_range_set(self, set, min, max, memo):
        # Iterate through the set allowing each item to build an element.
        children = [c.build_element(min, max, memo) for c in set]
        children = [c for c in children if c]

        # Wrap up results appropriately.
        if not children:
            return None
        if len(children) == 1:
            return children[0]
        else:
            return Alternative(children)


#---------------------------------------------------------------------------
# Element classes used in numeric grammar constructions.

class Collection(Compound):

    _element_name = "element"
    _default_value = None

    def __init__(self, spec, element, name=None):
        self._element = element
        Compound.__init__(self, spec, extras=[element], name=name)

    def value(self, node):
        child_node = node.get_child_by_name(self._element_name, shallow=True)

        if child_node:  return child_node.value()
        else:           return self._default_value


class Magnitude(Compound):

    _mul_default = 1
    _rem_default = 0

    def __init__(self, factor, spec, multiplier, remainder, name=None):
        self._factor = factor
        self._mul = multiplier
        self._rem = remainder
        Compound.__init__(self, spec, extras=[multiplier, remainder], name=name)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self._factor)

    def value(self, node):
        mul_node = node.get_child_by_name(self._mul.name, shallow=True)
        rem_node = node.get_child_by_name(self._rem.name, shallow=True)

        if mul_node:  multiplier = mul_node.value()
        else:         multiplier = self._mul_default
        if rem_node:  remainder  = rem_node.value()
        else:         remainder  = self._rem_default

        return multiplier * self._factor + remainder


class ModifiedPathsCollection(Alternative):

    # Collect and use paths to *augment* recognition of the given element
    #  (default).  Discards nil value or unchanged strings returned by the
    #  modifier function.
    MODE_AUGMENT = 1

    # Collect and use paths to *replace* recognition of the given element.
    #  Discards nil value strings returned by the modifier function.  Keeps
    #  unchanged path strings.
    MODE_REPLACE = 2

    def __init__(self, element, modifier_function, modifier_mode,
                 name=None):

        if not 0 < modifier_mode <= 2:
            raise ValueError("Invalid modifier mode: %d" % modifier_mode)

        # Generate element tree recognition paths and use them to create
        #  elements.
        children = []
        specs = set()
        for path in self._generate_paths(element):
            # Get a flat list of each word in this path.
            all_words = []
            for words in path:
                all_words.extend(words)

            # Get a spec using the modifier function.
            text = " ".join(all_words)
            spec = modifier_function(text)

            # Handle the result based on the specified modifier mode.
            if modifier_mode is self.MODE_AUGMENT:
                if not spec or text == spec:
                    continue
            elif modifier_mode is self.MODE_REPLACE:
                if not spec:
                    continue

            # Always skip duplicate specs.
            if spec in specs:
                continue

            # Initialize a new ModifiedPath element, passing the original
            #  element and words for decode-time calculation of the integer
            #  value.
            specs.add(spec)
            children.append(ModifiedPath(spec, all_words, element))

        # Initialize super class.
        Alternative.__init__(self, children=children, name=name)

    def _generate_sequence_paths(self, element, child_no):
        if child_no > len(element.children) - 1:
            yield []
            return

        child = element.children[child_no]
        next_child = child_no + 1
        for head in self._generate_paths(child):
            if None in head:
                continue
            for tail in self._generate_sequence_paths(element, next_child):
                yield head + tail

    def _generate_paths(self, element):
        if isinstance(element, Literal):
            yield [element.words]

        elif isinstance(element, ListRef):
            if len(element.list) == 0:
                # Impossible path.
                yield [None]
            else:
                # Generate paths from the list words.
                for word in element.list.get_list_items():
                    yield [word.split()]

        elif isinstance(element, Sequence):
            # Generate each complete path of the sequence.
            for path in self._generate_sequence_paths(element, 0):
                if None not in path:
                    yield path

        elif isinstance(element, Alternative):
            # Generate the paths of each alternative.
            for child in element.children:
                for path in self._generate_paths(child):
                    if None not in path:
                        yield path

        elif isinstance(element, Optional):
            # Generate the possible paths.
            # Optional element is not spoken.
            yield []

            # Optional element is spoken.
            for path in self._generate_paths(element.children[0]):
                if None not in path:
                    yield path

        elif isinstance(element, Impossible):
            # Impossible path.
            yield [None]

        elif isinstance(element, RuleRef):
            # Generate the possible paths from the referenced rule's element
            # tree.
            for path in self._generate_paths(element.rule.element):
                if None not in path:
                    yield path

        else:
            yield []


class ModifiedPath(Compound):

    def __init__(self, spec, words, orig_root_element, name=None):
        self._words = words
        self._orig_root_element = orig_root_element
        Compound.__init__(self, spec, name=name)

    def value(self, node):
        """
            The *value* of a :class:`ModifiedPath` is the *value*
            obtained by decoding the original words.

        """
        import dragonfly.grammar.state as state_
        words_rules = tuple((word, 0) for word in self._words)
        s = state_.State(words_rules, [], node.engine)
        s.initialize_decoding()
        for _ in self._orig_root_element.decode(s):
            if s.finished():
                root = s.build_parse_tree()
                return root.value()
        self._log_decode.error("CompoundWord %s: failed to decode original"
                               " words %r.", self, " ".join(self._words))
        return None


#---------------------------------------------------------------------------
# Integer content class.

class IntegerContentBase(object):
    builders = None



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

import sys

# --------------------------------------------------------------------------
from .config            import Config, Section, Item
from .error             import DragonflyError, GrammarError

# --------------------------------------------------------------------------
from .engines           import (get_engine, EngineError, MimicFailure,
                                get_current_engine)

# --------------------------------------------------------------------------
from .grammar.grammar_base       import Grammar
from .grammar.grammar_connection import ConnectionGrammar
from .grammar.rule_base          import Rule
from .grammar.rule_basic         import BasicRule
from .grammar.rule_compound      import CompoundRule
from .grammar.rule_mapping       import MappingRule
from .grammar.elements  import (ElementBase, Sequence, Alternative,
                                Optional, Repetition, Literal,
                                ListRef, DictListRef, Dictation, Modifier,
                                RuleRef, RuleWrap, Compound, Choice,
                                Empty, Impossible)

from .grammar.context   import Context, AppContext, FuncContext
from .grammar.list      import ListBase, List, DictList
from .grammar.recobs    import (RecognitionObserver, RecognitionHistory,
                                PlaybackHistory)
from .grammar.recobs_callbacks   import (CallbackRecognitionObserver,
                                         register_beginning_callback,
                                         register_recognition_callback,
                                         register_failure_callback,
                                         register_ending_callback,
                                         register_post_recognition_callback)

# --------------------------------------------------------------------------

from .actions           import (ActionBase, DynStrActionBase, ActionError,
                                Repeat, Key, Text, Mouse, Paste, Pause,
                                Mimic, Playback, WaitWindow, FocusWindow,
                                Function, StartApp, BringApp, PlaySound,
                                Typeable, Keyboard, typeables, RunCommand,
                                ContextAction)

if sys.platform.startswith("win"):
    from .actions       import (KeyboardInput, MouseInput, HardwareInput,
                                make_input_array, send_input_array)

# --------------------------------------------------------------------------

if sys.platform.startswith("win"):
    from .windows.clipboard import Clipboard
else:
    from .util              import Clipboard

# --------------------------------------------------------------------------

from .windows.rectangle import Rectangle, unit
from .windows.point     import Point
from .windows           import Window, Monitor, monitors

# --------------------------------------------------------------------------
from .language          import (Integer, IntegerRef, ShortIntegerRef,
                                Digits, DigitsRef,
                                Number, NumberRef)

# --------------------------------------------------------------------------
from .accessibility     import (CursorPosition, TextQuery,
                                get_accessibility_controller,
                                get_stopping_accessibility_controller)

from .rpc               import RPCServer, send_rpc_request



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Recognition observer classes
============================================================================

"""

import time
import logging

from six import integer_types

from ..actions.actions  import Playback
from ..engines          import get_engine


#---------------------------------------------------------------------------

class RecognitionObserver(object):
    """
    Recognition observer base class.

    Sub-classes should override one or more of the event methods.
    """

    _log = logging.getLogger("grammar")

    def __init__(self):
        pass

    def __del__(self):
        try:
            self.unregister()
        except Exception:
            pass

    def register(self):
        """
        Register the observer for recognition state events.
        """
        engine = get_engine()
        engine.register_recognition_observer(self)

    def unregister(self):
        """
        Unregister the observer for recognition state events.
        """
        engine = get_engine()
        engine.unregister_recognition_observer(self)

    def on_begin(self):
        """
        Method called when the observer is registered and speech start is
        detected.
        """

    def on_recognition(self, words, rule, node, results):
        """
        Method called when speech successfully decoded to a grammar rule or
        to dictation.

        This is called *before* grammar rule processing (i.e.
        ``Rule.process_recognition()``).

        :param words: recognized words
        :type words: tuple
        :param rule: *optional* recognized rule
        :type rule: Rule
        :param node: *optional* parse tree node
        :type node: Node
        :param results: *optional* engine recognition results object
        :type results: :ref:`engine-specific type<RefGrammarCallbackResultsTypes>`
        """

    def on_failure(self, results):
        """
        Method called when speech failed to decode to a grammar rule or to
        dictation.

        :param results: *optional* engine recognition results object
        :type results: :ref:`engine-specific type<RefGrammarCallbackResultsTypes>`
        """

    def on_end(self, results):
        """
        Method called when speech ends, either with a successful
        recognition (after ``on_recognition``) or in failure (after
        ``on_failure``).

        :param results: *optional* engine recognition results object
        :type results: :ref:`engine-specific type<RefGrammarCallbackResultsTypes>`
        """

    def on_post_recognition(self, words, rule, node, results):
        """
        Method called when speech successfully decoded to a grammar rule or
        to dictation.

        This is called *after* grammar rule processing (i.e.
        ``Rule.process_recognition()``).

        :param words: recognized words
        :type words: tuple
        :param rule: *optional* recognized rule
        :type rule: Rule
        :param node: *optional* parse tree node
        :type node: Node
        :param results: *optional* engine recognition results object
        :type results: :ref:`engine-specific type<RefGrammarCallbackResultsTypes>`
        """


#---------------------------------------------------------------------------

class RecognitionHistory(list, RecognitionObserver):
    """
        Storage class for recent recognitions.

        Instances of this class monitor recognitions and store them
        internally.  This class derives from the built in *list* type
        and can be accessed as if it were a normal *list* containing
        recent recognitions.  Note that an instance's contents are
        updated automatically as recognitions are received.

    """

    def __init__(self, length=10):
        list.__init__(self)
        RecognitionObserver.__init__(self)
        self._complete = True

        if (length is None or (isinstance(length, integer_types) and
                               length >= 1)):
            self._length = length
        else:
            raise ValueError("length must be a positive int or None,"
                             " received %r." % length)

    @property
    def complete(self):
        """ *False* if phrase-start detected but no recognition yet. """
        return self._complete

    def on_begin(self):
        """"""
        self._complete = False

    def on_recognition(self, words):
        """"""
        self._complete = True
        self.append(self._recognition_to_item(words))
        if self._length:
            while len(self) > self._length:
                self.pop(0)

    def _recognition_to_item(self, words):
        """"""
        # pylint: disable=no-self-use
        return words


#---------------------------------------------------------------------------

class PlaybackHistory(RecognitionHistory):
    """
        Storage class for playing back recent recognitions via the
        :class:`Playback` action.

        Instances of this class monitor recognitions and store them
        internally.  This class derives from the built in *list* type
        and can be accessed as if it were a normal *list* containing
        :class:`Playback` actions for recent recognitions.  Note that an
        instance's contents are updated automatically as recognitions are
        received.

    """

    def __init__(self, length=10):
        RecognitionHistory.__init__(self, length)

    def _recognition_to_item(self, words):
        return (words, time.time())

    def __getitem__(self, key):
        result = RecognitionHistory.__getitem__(self, key)
        if len(result) == 1:
            return Playback([(result[0][0], 0)])
        elif len(result) == 2 and isinstance(result[1], float):
            return Playback([(result[0], 0)])
        elif len(result) == 0:
            return Playback(())
        else:
            pairs = [(w1, t2 - t1)
                     for ((w1, t1), (w2, t2))
                     in zip(result[:-1], result[1:])]
            pairs.append((result[-1][0], 0))
            return Playback(pairs)

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))



#
# This file is part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
#   Dragonfly is free software: you can redistribute it and/or modify it
#   under the terms of the GNU Lesser General Public License as published
#   by the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Dragonfly is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with Dragonfly.  If not, see
#   <http://www.gnu.org/licenses/>.
#

"""
Dictation container for the Kaldi engine.

"""

import itertools

from six import text_type

from ...grammar.elements_basic import Dictation as BaseDictation
from ...grammar.elements_basic import ElementBase, RuleRef, Alternative, ListRef, DictListRef, Repetition
from ...grammar.list import List, DictList
from ...grammar.rule_compound import CompoundRule


#---------------------------------------------------------------------------
# User dictation class -- elements capable of also recognizing user words.

class UserDictation(RuleRef):
    """
        Imitates the standard Dictation element class, using individual chunks
        of Dictation or the user's added terminology.
    """

    # Partially copied from BaseDictation
    def __init__(self, name=None, format=True, default=None):
        RuleRef.__init__(self, rule=_user_dictation_sequence_rule, name=name, default=default)
        self._format_words = format
        self._string_methods = []

    # Use BaseDictation methods
    __repr__ = BaseDictation.__repr__
    __getattr__ = BaseDictation.__getattr__

    # Partially copied from BaseDictation
    def value(self, node):
        words = node.children[0].value()
        if self._format_words:
            return node.engine.DictationContainer(words, self._string_methods)
        else:
            return words

user_dictation_list = List('__kaldi_user_dictation_list', [])
user_dictation_dictlist = DictList('__kaldi_user_dictation_dictlist', {})

class _UserDictationSequenceRule(CompoundRule):
    spec = "<__kaldi_user_dictation_sequence>"
    extras = [
        Repetition(Alternative([
            BaseDictation(format=False),
            ListRef('__kaldi_user_dictation_listref', user_dictation_list),
            DictListRef('__kaldi_user_dictation_dictlistref', user_dictation_dictlist),
        ]), min=1, max=16, name="__kaldi_user_dictation_sequence"),
    ]
    exported = False
    def value(self, node):
        # This method returns the value of the root Repetition: a list of values of
        # Alternatives, each (dictation or listref or dictlistref) being a string.
        chunks = node.children[0].value()
        # Make sure each chunk is a list (of strings), then concat them together.
        return list(itertools.chain.from_iterable([chunk] if not isinstance(chunk, list) else chunk for chunk in chunks))

_user_dictation_sequence_rule = _UserDictationSequenceRule()


#---------------------------------------------------------------------------
# Alternative dictation classes -- elements capable of default or alternative dictation.

class AlternativeDictation(BaseDictation):

    alternative_default = True

    def __init__(self, *args, **kwargs):
        self.alternative = kwargs.pop('alternative', self.alternative_default)
        BaseDictation.__init__(self, *args, **kwargs)

class DefaultDictation(BaseDictation):

    alternative_default = False

    def __init__(self, *args, **kwargs):
        self.alternative = kwargs.pop('alternative', self.alternative_default)
        BaseDictation.__init__(self, *args, **kwargs)



"""
Tests for the Kaldi engine

Most engine functionality is tested here, although the tests are done
entirely via `mimic`, so there are some things which have to be tested
manually.

These tests assume that the `kaldi_model` model is used.

Adapted from `test_engine_sphinx.py`.
"""

import unittest

import logging

from dragonfly.engines import (EngineBase, get_engine)
from dragonfly.grammar.elements import Literal, Sequence
from dragonfly.test import ElementTester

try:
    from dragonfly.engines.backend_kaldi.engine import KaldiError
except ImportError:
    KaldiError = Exception


class MockLoggingHandler(logging.Handler):
    """
    Mock logging handler to check for expected logs.
    Adapted this from a Stack Overflow answer: https://stackoverflow.com/a/1049375
    """

    def __init__(self, *args, **kwargs):
        self.messages = {
            'debug': [],
            'info': [],
            'warning': [],
            'error': [],
            'critical': [],
        }
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        self.messages[record.levelname.lower()].append(record.getMessage())


class KaldiEngineTests(unittest.TestCase):
    """
    Tests for the Kaldi engine.
    """
    def test_get_engine_kaldi_is_usable(self):
        """
        Verify that the kaldi engine is usable by testing that a simple
        rule is loaded correctly and works correctly.
        """
        engine = get_engine()
        assert engine.name == "kaldi"
        assert isinstance(engine, EngineBase)
        engine.speak("testing kaldi")

        # Test that a basic rule can be loaded and recognized.
        seq = Sequence([Literal("hello"), Literal("world")])
        tester = ElementTester(seq, engine=engine)
        results = tester.recognize("hello world")
        self.assertEqual(results, [u"hello", u"world"])

    # FIXME: handling reseting user lexicon
    # def test_unknown_grammar_words(self):
    #     """ Verify that warnings are logged for a grammar with unknown words. """
    #     grammar = Grammar("test")
    #     grammar.add_rule(CompoundRule(name="r1", spec="testing unknownword"))
    #     grammar.add_rule(CompoundRule(name="r2", spec="wordz|morwordz"))

    #     # Catch log messages.
    #     handler = MockLoggingHandler()
    #     log = logging.getLogger("engine.compiler")
    #     log.addHandler(handler)
    #     grammar.load()
    #     log.removeHandler(handler)

    #     # Check the logged messages.
    #     warnings = handler.messages["warning"]
    #     self.assertEqual(len(warnings), 3)
    #     self.assertIn("unknownword", warnings[0])
    #     self.assertIn("wordz", warnings[1])
    #     self.assertIn("morwordz", warnings[2])
    #     self.assertNotIn("testing", '\n'.join(warnings))


# ---------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()
